# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import logging
from json import dump, dumps
from allensdk.config.model.description_parser import DescriptionParser
from allensdk.config.model.description import Description
from allensdk.config.model.formats.json_util import JsonUtil


class JsonDescriptionParser(DescriptionParser):
    log = logging.getLogger(__name__)
    
    
    def __init__(self):
        super(JsonDescriptionParser, self).__init__()
    
    
    def read(self, file_path, description=None, section=None, **kwargs):
        '''Parse a complete or partial configuration.
        
        Parameters
        ----------
        json_string : string
            Input to parse.
        description : Description, optional
            Where to put the parsed configuration.  If None a new one is created.
        section : string, optional
            Where to put the parsed configuration within the description.
        
        Returns
        -------
        Description
            The input description with parsed configuration added.
        
        Section is only specified for "bare" objects that are to be added to a section array.
        '''
        if description == None:
            description = Description()
            
        data = JsonUtil.read_json_file(file_path)
        description.unpack(data, section)
        
        return description
    
    
    def read_string(self, json_string, description=None, section=None, **kwargs):
        '''Parse a complete or partial configuration.
        
        Parameters
        ----------
        json_string : string
            Input to parse.
        description : Description, optional
            Where to put the parsed configuration.  If None a new one is created.
        section : string, optional
            Where to put the parsed configuration within the description.
        
        Returns
        -------
        Description
            The input description with parsed configuration added.
        
        Section is only specified for "bare" objects that are to be added to a section array.
        '''
        if description == None:
            description = Description()
        
        data = JsonUtil.read_json_string(json_string)
        
        description.unpack(data, section)
        
        return description
    
    
    def write(self, filename, description):
        '''Write the description to a JSON file.
        
        Parameters
        ----------
        description : Description
            Object to write.
        '''
        try:
            with open(filename, 'w') as f:
                    dump(description.data, f, indent=2)

        except Exception:
            self.log.warn("Couldn't write allensdk json description: %s" % filename)
            raise
        
        return
    
    
    def write_string(self, description):
        '''Write the description to a JSON string.
        
        Parameters
        ----------
        description : Description
            Object to write.
        
        Returns
        -------
        string
           JSON serialization of the input.
        '''
        try:
            json_string = dumps(description.data,
                                indent=2)
            return json_string
        except Exception:
            self.log.warn("Couldn't write allensdk json description: %s")
            raise



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

""" The methods in this module are used for configuring dynamics and reset rules for the GlifNeuron.  
For more details on how to use these methods, see :doc:`glif_models`.
"""

import functools
import numpy as np


class GlifNeuronMethod( object ):
    """ A simple class to keep track of the name and parameters associated with a neuron method.
    This class is initialized with a name, function, and parameters to pass to the function.  The
    function then has those passed parameters fixed to a partial function using functools.partial.  
    This class then mimics a function itself using the __call__ convention. Parameters that are not 
    fixed in this way are assumed to be passed into the method when it is called.  If the passed
    parameters contain an argument that is not part of the function signature, an exception will
    be raised.    

    Parameters
    ----------
    method_name : string
        A shorthand name that will be used to reference this method in the `GlifNeuron`.
    method : function
        A python function to be called when this instance is called.
    method_params : dict
        A dictionary mapping function arguments to values for values that should be fixed.
    """

    def __init__(self, method_name, method, method_params):
        self.name = method_name
        self.params = method_params
        self.method = functools.partial(method, **method_params)

    def __call__(self, *args, **kwargs):
        """ Defining this method allows an instance to be called like a function """
        return self.method(*args, **kwargs)

    def to_dict(self):
        return {
            'name': self.name,
            'params': self.params
        }

    def modify_parameter(self, param, operator):
        """ Modify a function parameter that needs to be modified after initialization. 

        Parameters
        ----------
        param : string
            the name of the parameter to modify
        operator : callable
            a function or lambda that returns the desired modified value

        Returns
        -------
        type
            the new value of the variable that was just modified.
        """
        value = operator(self.method.keywords[param])
        self.method.keywords[param] = value
        return value


def dynamics_AScurrent_exp(neuron, AScurrents_t0, time_step, spike_time_steps):
    """ Exponential afterspike current dynamics method takes a current at t0 and returns the current at
    a time step later.
    
    Parameters
    ----------
    AScurrents_t0: np.ndarray
        value of the after-spike currents 
    time_step: int
        index of global time step
    spike_time_steps: list
        index of global time of all model spikes

    Returns
    -------
    type: np.ndarray
        the new value of after-spike currents.

    """
    return AScurrents_t0*(1.0 + neuron.k*neuron.dt)
    

def dynamics_AScurrent_none(neuron, AScurrents_t0, time_step, spike_time_steps):
    """ This method always returns zeros for the afterspike currents, regardless of input. 
    
    Parameters
    ----------
    AScurrents_t0: np.ndarray
        value of the after-spike currents 
    time_step: int
        index of global time step
    spike_time_steps: list
        index of global time of all model spikes


    Returns
    -------
    type: np.ndarray
        array of zeros for the after-spike currents.

    """
    return np.zeros(len(AScurrents_t0))


def dynamics_voltage_forward_euler(neuron, voltage_t0, AScurrents_t0, inj):
    """ Linear voltage dynamics simulated by standard forward Euler. 
    
    Parameters
    ----------
    voltage_t0: float
        value of voltage
    AScurrents_t0: np.ndarray
        value of the after-spike currents 
    inj: float
        value of stimulus current being injected

    Returns
    -------
    type: float
        value of voltage at the next time step.

    """
    return voltage_t0 + (inj + np.sum(AScurrents_t0) - neuron.G * neuron.coeffs['G'] * (voltage_t0 - neuron.El)) * neuron.dt / (neuron.C * neuron.coeffs['C'])

def dynamics_voltage_euler_exact(neuron, voltage_t0, AScurrents_t0, inj):
    """ Linear voltage dynamics simulated by the exponential kernel to calculate the exact voltage values at the time step.
        This assumes constant current (calculated as an average) over the time step. 
    
    Parameters
    ----------
    voltage_t0: float
        value of voltage
    AScurrents_t0: np.ndarray
        value of the after-spike currents 
    inj: float
        value of stimulus current being injected

    Returns
    -------
    type: float
        value of voltage at the next time step

    """

    tau = (neuron.C * neuron.coeffs['C'])
    I = inj + np.sum(AScurrents_t0)
    g = neuron.G * neuron.coeffs['G']
    A = g/tau
    N = (I+ g*neuron.El)/tau

    return voltage_t0*np.exp(-neuron.dt*A) + N*(1-np.exp(-A*neuron.dt))/A
    

def dynamics_threshold_three_components(neuron, threshold_t0, voltage_t0):        
    """The threshold will adapt via two mechanisms: 1. a slower voltage dependent adaptation as in the dynamics_threshold_adapt_standard. 
    These are the components which are fit via optimization and inditial conditions are supplied via the GLM. 2. A fast component initiated 
    by a spike which quickly decays.  These values are estimated via the tri short square stimuli. 
    
    Parameters
    ----------
    threshold_t0: float
        value of threshold
    voltage_t0: float
        value of voltage

    Returns
    -------
    type: float
        value of overall threshold at the next time step.

    """
    
    # update_method_data keeps track of parameters that need to be shared between methods
    md = neuron.update_method_data

    # initial conditions
    if 'th_spike' not in md:
        md['th_spike'] = [ 0 ]
    if 'th_voltage' not in md:
        md['th_voltage'] = [ 0 ]

    th_spike = md['th_spike'][-1]
    th_voltage = md['th_voltage'][-1] 

    voltage_component = th_voltage + ( md['a_voltage'] *  ( voltage_t0 - neuron.El ) - 
                                       md['b_voltage'] * neuron.coeffs['b'] * ( th_voltage ) ) * neuron.dt
    spike_component = th_spike - md['b_spike'] * th_spike * neuron.dt

    # update the subtrheshold voltage and spiking values of the threshold
    md['th_spike'].append(spike_component)
    md['th_voltage'].append(voltage_component)
    
    return neuron.coeffs['a'] * voltage_component + spike_component+neuron.th_inf * neuron.coeffs['th_inf']    
    

def dynamics_threshold_inf(neuron, threshold_t0, voltage_t0):
    """ Set threshold to the neuron's instantaneous threshold.
    
    Parameters
    ----------
    threshold_t0: float
        value of threshold
    voltage_t0: float
        value of voltage

    Returns
    -------
    type: float
        value of overall threshold at the next time step.
    """
    return neuron.coeffs['th_inf'] * neuron.th_inf
    

def reset_AScurrent_sum(neuron, AScurrents_t0, r):
    """ Reset afterspike currents by adding summed exponentials. Left over currents from last spikes as 
    well as newly initiated currents from current spike.  Currents amplitudes in neuron.asc_vector need
    to be the amplitudes advanced though the spike cutting.  I.e. In the preprocessor if the after spike currents 
    are calculated via the GLM from spike initiation the amplitude at the time after the spike cutting needs to 
    be calculated and neuron.asc_vector needs to be set to this value.
    
    Parameters
    ----------
    r : np.ndarray
        a coefficient vector applied to the afterspike currents
    AScurrents_t0: np.ndarray
        value of the after-spike currents 

    Returns
    -------
    type: np.ndarray
        array of new after-spike currents.
    """

    # neuron.asc_vector are amplitudes initiating after the spike is cut
    new_currents = neuron.asc_vector * neuron.coeffs['asc_vector'] 

    #advancing cut currents though the spike    
    left_over_currents=AScurrents_t0 * r * np.exp(neuron.k * neuron.dt * neuron.spike_cut_length) 

    return new_currents+left_over_currents


def reset_AScurrent_none(neuron, AScurrents_t0):
    """ Reset afterspike currents to zero. 

    Parameters
    ----------
    AScurrents_t0: np.ndarray
        value of the after-spike currents 

    Returns
    -------
    type: np.ndarray
        array of zeros for the after-spike currents.

    """   
    if np.sum(AScurrents_t0)!=0:
        raise Exception('You are running a LIF but the AScurrents are not zero!')
    return np.zeros(len(AScurrents_t0))


def reset_voltage_bio_rules(neuron, voltage_t0, a, b):
    """ Reset voltage to the previous value with a scale and offset applied derived from the electrophysiological data.

    Parameters
    ----------
    voltage_t0: float
        value of voltage at spike
    a : float
        voltage scale constant
    b : float
        voltage offset constant

    Returns
    -------
    type: float
        value of voltage after the spike
    """

    return a*(voltage_t0)+b


def reset_voltage_zero(neuron, voltage_t0):
    """ Reset voltage to zero. 
    
    Parameters
    ----------
    voltage_t0: float
        value of voltage at spike

    Returns
    -------
    type: float
        zero value of voltage after the spike
    """

    return 0.0


def reset_threshold_inf(neuron, threshold_t0, voltage_v1):
    """ Reset the threshold to instantaneous threshold. 
    
    Parameters
    ----------
    threshold_t0: float
        value of threshold at spike
    voltage_v1: float
        voltage after reset

    Returns
    -------
    type: float
        value of threshold after the spike
    """
    return neuron.coeffs['th_inf'] * neuron.th_inf


def reset_threshold_three_components(neuron, threshold_t0, voltage_v1):
    """This method resets the two components of the threshold: a spike
    component and a subthreshold voltage component which are added to the instantaneous threshold.
    
    Parameters
    ----------
    threshold_t0: float
        value of threshold at spike
    voltage_v1: float
        voltage after reset

    Returns
    -------
    type: float
        value of threshold after the spike
    """
    
    # update_method_data keeps track of parameters that need to be shared between methods
    md = neuron.update_method_data
    
    # update the subthreshold voltage and spiking values of the threshold
    md['th_spike'].append( md['th_spike'][-1] + md['a_spike'] )
    md['th_voltage'].append( md['th_voltage'][-1] ) #note these are the same value.

    return md['th_spike'][-1] + md['th_voltage'][-1] + neuron.th_inf * neuron.coeffs['th_inf']


#: The METHOD_LIBRARY constant groups dynamics and reset methods by group name (e.g. 'voltage_dynamics_method').  
#Those groups assign each method in this file a string name.  This is used by the GlifNeuron when initializing its dynamics and reset methods.
METHOD_LIBRARY = {
    'AScurrent_dynamics_method': { 
        'exp_glm': dynamics_AScurrent_exp,
        'none': dynamics_AScurrent_none
        },
    'voltage_dynamics_method': { 
        #Note that both of these functions simulate linear voltage dynamics. 
        #the "linear" uses first order (i.e. forward) Euler for simulation while the "euler_exact" 
        #uses the exponential kernel to calculate the exact voltage values at the time step.
        # This assumes constant current (calculated as an average) over the time step. 
        'linear': dynamics_voltage_forward_euler,
        'euler_exact':dynamics_voltage_euler_exact
        },
    'threshold_dynamics_method': {
        'adapt_sum_slow_fast':  dynamics_threshold_three_components,                        
        'adapt_rebound': dynamics_threshold_three_components,
        'inf': dynamics_threshold_inf
        },
    'AScurrent_reset_method': {
        'sum': reset_AScurrent_sum,
        'none': reset_AScurrent_none
        }, 
    'voltage_reset_method': {
        'v_before': reset_voltage_bio_rules,
        'zero': reset_voltage_zero,
        }, 
    'threshold_reset_method': {
        'inf': reset_threshold_inf,
        'adapt_sum_slow_fast': reset_threshold_three_components
        }
}



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.api.api import Api
import os, json
from collections import OrderedDict

class BiophysicalPerisomaticApi(Api):
    _NWB_file_type = 'NWB'
    _SWC_file_type = '3DNeuronReconstruction'
    _MOD_file_type = 'BiophysicalModelDescription'
    _FIT_file_type = 'NeuronalModelParameters'
    
    def __init__(self, base_uri=None):
        super(BiophysicalPerisomaticApi, self).__init__(base_uri)
        self.cache_stimulus = True
        self.ids = {}
        self.sweeps = []
        self.manifest = {}
    
    
    def build_rma(self, neuronal_model_id, fmt='json'):
        '''Construct a query to find all files related to a neuronal model.
        
        Parameters
        ----------
        neuronal_model_id : integer or string representation
            key of experiment to retrieve.
        fmt : string, optional
            json (default) or xml
        
        Returns
        -------
        string
            RMA query url.
        '''
        include_associations = ''.join([
            'neuronal_model_template(well_known_files(well_known_file_type)),',
            'specimen',
            '(ephys_result(well_known_files(well_known_file_type)),'
            'neuron_reconstructions(well_known_files(well_known_file_type)),',
            'ephys_sweeps),',
            'well_known_files(well_known_file_type)'])
        criteria_associations = ''.join([
            ("[id$eq%d]," % (neuronal_model_id)),
            include_associations])
        
        return ''.join([self.rma_endpoint, 
                        '/query.',
                        fmt,
                        '?q=',
                        'model::NeuronalModel,',
                        'rma::criteria,',
                        criteria_associations,
                        ',rma::include,',
                        include_associations])
    
    
    def read_json(self, json_parsed_data):
        '''Get the list of well_known_file ids from a response body
        containing nested sample,microarray_slides,well_known_files.
        
        Parameters
        ----------
        json_parsed_data : dict
           Response from the Allen Institute Api RMA.
        
        Returns
        -------
        list of strings
            Well known file ids.
        '''
        self.ids = {
            'stimulus': {},
            'morphology': {},
            'modfiles': {},
            'fit': {}
        }
        self.sweeps = []
        
        if 'msg' in json_parsed_data:
            for neuronal_model in json_parsed_data['msg']:
                if 'well_known_files' in neuronal_model:
                    for well_known_file in neuronal_model['well_known_files']:
                        if ('id' in well_known_file and
                            'path' in well_known_file and
                            self.is_well_known_file_type(well_known_file,
                                                         BiophysicalPerisomaticApi._FIT_file_type)):
                            self.ids['fit'][str(well_known_file['id'])] = \
                                os.path.split(well_known_file['path'])[1]
                
                if 'neuronal_model_template' in neuronal_model:
                    neuronal_model_template = neuronal_model['neuronal_model_template']
                    if 'well_known_files' in neuronal_model_template:
                        for well_known_file in neuronal_model_template['well_known_files']:
                            if ('id' in well_known_file and
                                'path' in well_known_file and
                                self.is_well_known_file_type(well_known_file,
                                                             BiophysicalPerisomaticApi._MOD_file_type)):
                                self.ids['modfiles'][str(well_known_file['id'])] = \
                                    os.path.join('modfiles',
                                                 os.path.split(well_known_file['path'])[1])
                
                if 'specimen' in neuronal_model:
                    specimen = neuronal_model['specimen']
                    
                    if 'neuron_reconstructions' in specimen:
                        for neuron_reconstruction in specimen['neuron_reconstructions']:
                            if 'well_known_files' in neuron_reconstruction:
                                for well_known_file in neuron_reconstruction['well_known_files']:
                                    if ('id' in well_known_file and
                                        'path' in well_known_file and
                                        self.is_well_known_file_type(well_known_file,
                                                                     BiophysicalPerisomaticApi._SWC_file_type)):
                                        self.ids['morphology'][str(well_known_file['id'])] = \
                                            os.path.split(well_known_file['path'])[1]
                    
                    if 'ephys_result' in specimen:
                        ephys_result = specimen['ephys_result']
                        if 'well_known_files' in ephys_result:
                            for well_known_file in ephys_result['well_known_files']:
                                if ('id' in well_known_file and
                                    'path' in well_known_file and
                                    self.is_well_known_file_type(well_known_file,
                                                                 BiophysicalPerisomaticApi._NWB_file_type)):
                                        self.ids['stimulus'][str(well_known_file['id'])] = \
                                            "%d.nwb" % (ephys_result['id'])
                    
                    
                    self.sweeps = [sweep['sweep_number'] 
                                   for sweep in specimen['ephys_sweeps']
                                   if sweep['stimulus_name'] != 'Test']
        
        return self.ids
    
    
    def is_well_known_file_type(self, wkf, name):
        '''Check if a structure has the expected name.
        
        Parameters
        ----------
        wkf : dict
            A well-known-file structure with nested type information.
        name : string
            The expected type name
        
        See Also
        --------
        read_json: where this helper function is used.
        '''
        try:
            return wkf['well_known_file_type']['name'] == name
        except:
            return False
    
    
    def get_well_known_file_ids(self, neuronal_model_id):
        '''Query the current RMA endpoint with a neuronal_model id
        to get the corresponding well known file ids.
        
        Returns
        -------
        list
            A list of well known file id strings.
        '''
        rma_builder_fn = self.build_rma
        json_traversal_fn = self.read_json
        
        return self.do_rma_query(rma_builder_fn, json_traversal_fn, neuronal_model_id)
    
    
    def create_manifest(self,
                        fit_path='',
                        stimulus_filename='',
                        swc_morphology_path='',
                        sweeps=[]):
        '''Generate a json configuration file with parameters for a 
        a biophysical experiment.
        
        Parameters
        ----------
        fit_path : string
            filename of a json configuration file with cell parameters.
        stimulus_filename : string
            path to an NWB file with input currents.
        swc_morphology_path : string
            file in SWC format.
        sweeps : array of integers
            which sweeps in the stimulus file are to be used.
        '''
        self.manifest = OrderedDict()
        self.manifest['biophys'] = [{
                'model_file': [ 'manifest.json',  fit_path ]
            }]
        self.manifest['runs'] = [{
                'sweeps': sweeps
            }]
        self.manifest['neuron'] = [{
                'hoc': [ 'stdgui.hoc', 'import3d.hoc' ]
            }]
        self.manifest['manifest'] = [
                {
                    'type': 'dir',
                    'spec': '.',
                    'key': 'BASEDIR'
                },
                {
                    'type': 'dir',
                    'spec': 'work',
                    'key': 'WORKDIR',
                    'parent': 'BASEDIR'
                },
                {
                    'type': 'file',
                    'spec': swc_morphology_path,
                    'key': 'MORPHOLOGY'
                },
                {
                    'type': 'dir',
                    'spec': 'modfiles',
                    'key': 'MODFILE_DIR'
                },
                {
                    'type': 'file',
                    'format': 'NWB',
                    'spec': stimulus_filename,
                    'key': 'stimulus_path'
                },
                {
                  'parent_key': 'WORKDIR', 
                  'type': 'file', 
                  'format': 'NWB',
                  'spec': stimulus_filename, 
                  'key': 'output'
                }
            ]
    
    
    def cache_data(self,
                   neuronal_model_id,
                   working_directory=None):
        '''Take a an experiment id, query the Api RMA to get well-known-files
        download the files, and store them in the working directory.
        
        Parameters
        ----------
        neuronal_model_id : int or string representation
            found in the neuronal_model table in the api
        working_directory : string
            Absolute path name where the downloaded well-known files will be stored.
        '''
        if working_directory is None:
            working_directory = self.default_working_directory
        try:
            os.stat(working_directory)
        except:
            os.mkdir(working_directory)
        
        
        work_dir = os.path.join(working_directory, 'work')
        try:
            os.stat(work_dir)
        except:
            os.mkdir(work_dir)
        
        modfile_dir = os.path.join(working_directory, 'modfiles')
        try:
            os.stat(modfile_dir)
        except:
            os.mkdir(modfile_dir)
        
        well_known_file_id_dict = self.get_well_known_file_ids(neuronal_model_id)
        
        for key, id_dict in well_known_file_id_dict.items():
            if (not self.cache_stimulus) and (key == 'stimulus'):
                continue
            
            for well_known_id, filename in id_dict.items():
                well_known_file_url = self.construct_well_known_file_download_url(well_known_id)
                cached_file_path = os.path.join(working_directory, filename)
                self.retrieve_file_over_http(well_known_file_url, cached_file_path)
        
        fit_path = self.ids['fit'].values()[0]
        stimulus_filename = self.ids['stimulus'].values()[0]
        swc_morphology_path = self.ids['morphology'].values()[0]
        sweeps = sorted(self.sweeps)
        
        self.create_manifest(fit_path,
                             stimulus_filename,
                             swc_morphology_path,
                             sweeps)
        
        manifest_path = os.path.join(working_directory, 'manifest.json')
        with open(manifest_path, 'wb') as f:
            f.write(json.dumps(self.manifest, indent=2))



# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import logging
from pprint import pprint, pformat
from allensdk.config.model.description import Description
from allensdk.config.model.description_parser import DescriptionParser


class PycfgDescriptionParser(DescriptionParser):
    log = logging.getLogger(__name__)
    
    def __init__(self):
        super(PycfgDescriptionParser, self).__init__()
    
    
    def read(self, pycfg_file_path, description=None, section=None, **kwargs):
        '''Read a serialized description from a Python (.pycfg) file.
        
        Parameters
        ----------
        filename : string
            Name of the .pycfg file.
        
        Returns
        -------
        Description
            Configuration object.
        '''
        header = kwargs.get('prefix', '')
        
        with open(pycfg_file_path, 'r') as f:
            return self.read_string(f.read(), description, section, header=header)
    
    
    def read_string(self, python_string, description=None, section=None, **kwargs):
        '''Read a serialized description from a Python (.pycfg) string.
        
        Parameters
        ----------
        python_string : string
            Python string with a serialized description.
        
        Returns
        -------
        Description
            Configuration object.
        '''
        
        if description == None:
            description = Description()
        
        header = kwargs.get('header', '')
                    
        python_string = "%s\n\nallensdk_description = %s" % (header, python_string)
        
        ns = {}
        code = compile(python_string, 'string', 'exec')
        exec(code, ns)
        data = ns['allensdk_description']
        description.unpack(data, section)
        
        return description


    def write(self, filename, description):
        '''Write the description to a Python (.pycfg) file.
        
        Parameters
        ----------
        filename : string
            Name of the file to write.
        '''
        try:
            with open(filename, 'w') as f:
                    pprint(description.data, f, indent=2)

        except Exception:
            self.log.warn("Couldn't write allensdk python description: %s" % filename)
            raise
        
        return
    
    def write_string(self, description):
        '''Write the description to a pretty-printed Python string.
        
        Parameters
        ----------
        description : Description
            Configuration object to write.
        '''
        pycfg_string = pformat(description.data, indent=2)
        
        return pycfg_string
        


# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.api.api import Api

import os, json, logging

class GlifApi(Api):
    def __init__(self, base_uri=None):
        super(GlifApi, self).__init__(base_uri)

        self.neuronal_model = None
        self.ephys_sweeps = None
        self.stimulus_url = None
        self.neuron_config_url = None

    def list_neuronal_models(self):
        ''' Query the API for a list of all GLIF neuronal models. 

        Returns
        -------
        list
            Meta data for all GLIF neuronal models.
        '''
        
        return self.do_rma_query(self.build_neuronal_model_list_rma_url, self.read_neuronal_model_list_json)


    def build_neuronal_model_list_rma_url(self, fmt='json'):
        include_associations = "specimen(ephys_result),neuronal_model_template[name$il'*LIF*']"

        url = ''.join([self.rma_endpoint, 
                       '/query.',
                       fmt,
                       '?q=',
                       'model::NeuronalModel,',
                       'rma::include,',
                       include_associations,
                       ',rma::options[num_rows$eq2000]'])

        return url
        

    def read_neuronal_model_list_json(self, json_parsed_data):
        ''' Return basic meta data for all neuronal models'''
        return json_parsed_data['msg']


    def get_neuronal_model(self, neuronal_model_id):
        '''Query the current RMA endpoint with a neuronal_model id
        to get the corresponding well known files and meta data.
        
        Returns
        -------
        dict
            A dictionary containing 
        '''
        
        self.metadata = self.do_rma_query(self.build_neuronal_model_rma_url, self.read_neuronal_model_json, neuronal_model_id)
        return self.metadata


    def get_ephys_sweeps(self):
        ''' Retrieve ephys sweep information out of downloaded metadata for a neuronal model 
        
        Returns
        -------
        list
            A list of sweeps metadata dictionaries
        '''
        return self.ephys_sweeps


    def get_neuron_config(self, output_file_name=None):
        ''' Retrieve a model configuration file from the API, optionally save it to disk, and 
        return the contents of that file as a dictionary.

        Parameters
        ----------
        output_file_name: string
            File name to store the neuron configuration (optional).
        '''
        if self.neuron_config_url is None:
            raise Exception("URL for neuron config file is empty.")

        neuron_config = self.retrieve_parsed_json_over_http(self.api_url + self.neuron_config_url)

        if output_file_name:
            with open(output_file_name, 'wb') as f:
                f.write(json.dumps(neuron_config, indent=2))
        
        return neuron_config

    def cache_stimulus_file(self, output_file_name):
        ''' Download the NWB file for the current neuronal model and save it to a file.

        Parameters
        ----------
        output_file_name: string
            File name to store the NWB file.
        '''
        if self.stimulus_url is None:
            raise Exception("URL for stimulus file is empty.")

        self.retrieve_file_over_http(self.api_url + self.metadata['stimulus_url'], output_file_name)


    def build_neuronal_model_rma_url(self, neuronal_model_id, fmt='json'):
        '''Construct a query to find all files related to a GLIF neuronal model.
        
        Parameters
        ----------
        neuronal_model_id : integer or string representation
            key of experiment to retrieve.
            
        Returns
        -------
        string
            RMA query url.
        '''
        
        include_associations = ''.join([
                'neuronal_model_template(well_known_files(well_known_file_type)),',
                'specimen(ephys_sweeps,ephys_result(well_known_files(well_known_file_type))),',
                'well_known_files(well_known_file_type)'])
                
        criteria_associations = ''.join([
                ("[id$eq%d]" % (neuronal_model_id))])
        
        url = ''.join([self.rma_endpoint, 
                       '/query.',
                       fmt,
                       '?q=',
                       'model::NeuronalModel,',
                       'rma::criteria,',
                       criteria_associations,
                       ',rma::include,',
                       include_associations])

        return url


    def read_neuronal_model_json(self, json_parsed_data):
        ''' Reformat the RMA query results into a more usable dictionary
        
        Parameters
        ----------
        json_parsed_data: dict
            the json response from the Allen Institute API RMA.

        Returns
        -------
        hash
            a dictionary containing fields necessary to run a GLIF model
        '''

        self.ephys_sweeps = None
        self.neuron_config_url = None
        self.stimulus_url = None            
        self.neuronal_model = json_parsed_data['msg'][0]

        # sweeps come from the specimen
        try:
            specimen = self.neuronal_model['specimen']
            self.ephys_sweeps = specimen['ephys_sweeps']
        except Exception, e:
            self.ephys_sweeps = None
            
        if self.ephys_sweeps is None:
            logging.warning("Could not find ephys_sweeps for this model (%d)" % self.neuronal_model['id'])

        
        # neuron config file comes from the neuronal model's well known files
        try:
            for wkf in self.neuronal_model['well_known_files']:
                if wkf['path'].endswith('neuron_config.json'):
                    self.neuron_config_url = wkf['download_link']
                    break
        except Exception, e:
            self.neuron_config_url = None

        if self.neuron_config_url is None:
            logging.warning("Could not find neuron config well_known_file for this model (%d)" % self.neuronal_model['id'])

        # NWB file comes from the ephys_result's well known files
        try:
            ephys_result = specimen['ephys_result']
            for wkf in ephys_result['well_known_files']:
                if wkf['well_known_file_type']['name'] == 'NWB':
                    self.stimulus_url = wkf['download_link']
                    break
        except Exception, e:
            self.stimulus_url = None

        if self.stimulus_url is None:
            logging.warning("Could not find stimulus well_known_file for this model (%d)" % self.neuronal_model['id'])


        data = {
            'neuron_config_url': self.neuron_config_url,
            'stimulus_url': self.stimulus_url,
            'ephys_sweeps': self.ephys_sweeps,
            'neuronal_model': self.neuronal_model
            }

        return data



    
        



# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import logging

class Manifest(object):
    """Manages the location of external files 
     referenced in an Allen SDK configuration """

    DIR = 'dir'
    FILE = 'file'
    DIRNAME = 'dir_name'
    log = logging.getLogger(__name__)

    def __init__(self, config=None):
        self.path_info = {}
        
        if config != None:
            self.load_config(config)
       
        
    def load_config(self, config):
        ''' Load paths into the manifest from an Allen SDK config section.
        
        Parameters
        ----------
        config : Config
            Manifest section of an Allen SDK config.
        '''
        for path_info in config:
            path_type = path_info['type']
            path_format = None
            if 'format' in path_info:
                path_format = path_info['format']
            
            if path_type == 'file':
                try:
                    parent_key = path_info['parent_key']
                except:
                    parent_key = None
                
                self.add_file(path_info['key'],
                              path_info['spec'],
                              parent_key,
                              path_format)
            elif path_type == 'dir':
                try:
                    parent_key = path_info['parent_key']
                except:
                    parent_key = None
                
                spec = path_info['spec']
                absolute = False
                if spec[0] == '/':
                    absolute = True
                self.add_path(path_info['key'],
                              path_info['spec'],
                              path_type,
                              absolute,
                              path_format,
                              parent_key)
            else:
                Manifest.log.warning("Unknown path type in manifest: %s" %
                                     (path_type))
                        
        
    def add_path(self, key, path, path_type=DIR,
                 absolute=True, path_format=None, parent_key=None):
        '''Insert a new entry.
        
        Parameters
        ----------
        key : string
            Identifier for referencing the entry.
        path : string
            Specification for a path using %s, %d style substitution.
        path_type : string enumeration
            'dir' (default) or 'file'
        absolute : boolean
            Is the spec relative to the process current directory.
        path_format : string, optional
            Indicate a known file type for further parsing.
        parent_key : string
            Refer to another entry.
        '''
        if parent_key:
            path_args = []
            
            try:
                parent_path = self.path_info[parent_key]['spec']
                path_args.append(parent_path)
            except:
                Manifest.log.error("cannot resolve directory key %s" % (parent_key))
                raise
            path_args.extend(path.split('/'))
            path = os.path.join(*path_args)
        
        # TODO: relative paths need to be considered better
        if absolute == True:
            path = os.path.abspath(path)
        else:
            path = os.path.abspath(path)
            
        if path_type == Manifest.DIRNAME:
            path = os.path.dirname(path)
            
        self.path_info[key] = { 'type': path_type,
                                'spec': path}
        
        if path_type == Manifest.FILE and path_format != None:
            self.path_info[key]['format'] = path_format


    def add_paths(self, path_info):
        ''' add information about paths stored in the manifest.
        
        Parameters
            path_info : dict
                Information about the new paths
        '''
        for path_key, path_data in path_info.items():
            path_format = None
            
            if 'format' in path_data:
                path_format = path_data['format']
            
            Manifest.log.info("Adding path.  type: %s, format: %s, spec: %s" %
                              (path_data['type'],
                               path_data['spec'],
                               path_format))
            entry = { 'type': path_data['type'],
                      'spec': path_data['spec'] 
                    }
            if path_format != None:
                entry['format'] = path_format

            self.path_info[path_key] = entry
    
    
    def add_file(self,
                 file_key,
                 file_name,
                 dir_key=None,
                 path_format=None):
        '''Insert a new file entry.
        
        Parameters
        ----------
        file_key : string
            Reference to the entry.
        file_name : string
            Subtitutions of the %s, %d style allowed.
        dir_key : string
            Reference to the parent directory entry.
        path_format : string, optional
            File type for further parsing.
        '''
        path_args = []
        
        if dir_key:
            try:
                dir_path = self.path_info[dir_key]['spec']
                path_args.append(dir_path)
            except:
                Manifest.log.error("cannot resolve directory key %s" % (dir_key))
                raise
        elif not file_name.startswith('/'):
            path_args.append(os.curdir)
        else:
            path_args.append(os.path.sep)
        
        path_args.extend(file_name.split('/'))
        file_path = os.path.join(*path_args)
        
        self.path_info[file_key] = { 'type': Manifest.FILE,
                                     'spec': file_path }
        
        if path_format:
            self.path_info[file_key]['format'] = path_format
    
    
    def get_path(self, path_key, *args):
        '''Retrieve an entry with substitutions.
        
        Parameters
        ----------
        path_key : string
            Refer to the entry to retrieve.
        args : any types, optional
           arguments to be substituted into the path spec for %s, %d, etc.
        
        Returns
        -------
        string
            Path with parent structure and substitutions applied.
        '''
        path_spec = str(self.path_info[path_key]['spec'].encode('ascii',
                                                                'ignore'))
        
        if args != None and len(args) != 0:
            path = path_spec % args
        else:
            path = path_spec
            
        return path
    
    
    def get_format(self, path_key):
        '''Retrieve the type of a path entry.
        
        Parameters
        ----------
        path_key : string
            reference to the entry
        
        Returns
        -------
        string
            File type.
        '''
        path_entry = self.path_info[path_key]
        path_format = None
        
        if 'format' in path_entry:
            path_format = path_entry['format']
            
        return path_format
    
    
    def create_dir(self, path_key):
        '''Make a directory for an entry.
        
        Parameters
        ----------
        path_key : string
            Reference to the entry.
        '''
        dir_path = self.get_path(path_key)
        
        try:
            os.stat(dir_path)
        except:
            os.mkdir(dir_path)
    
    
    def check_dir(self, path_key, do_exit=False):
        '''Verify a directories existence or optionally exit.
        
        Parameters
        ----------
        path_key : string
            Reference to the entry.
        do_exit : boolean
            What to do if the directory is not present.
        '''
        dir_path = self.get_path(path_key)
        
        if not os.path.exists(dir_path):
            Manifest.log.fatal('Directory %s does not exist; exiting.' % 
                               (dir_path))
            if do_exit == True:
                quit()
    
    def resolve_paths(self, description_dict, suffix='_key'):
        '''Walk input items and expand those that refer to a manifest entry.
        
        Parameters
        ----------
        description_dict : dict
            Any entries with key names ending in suffix will be expanded.
        suffix : string
            Indicates the entries to be expanded.
        '''
        key_pattern =  re.compile('(.*)%s$' % (suffix))
        
        for description_key, manifest_key in description_dict.items():
            m = key_pattern.match(description_key)
            if m:
                real_key = m.group(1)  # i.e. job_dir_key -> job_dir
                filename = self.get_path(manifest_key)
                description_dict[real_key] = filename
                del description_dict[description_key]



# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import logging
from allensdk.config.model.description import Description


class DescriptionParser(object):
    log = logging.getLogger(__name__)
    
    
    def __init__(self):
        pass
    
    
    def read(self, file_path, description=None, section=None, **kwargs):
        '''Parse data needed for a simulation.
        
        Parameters
        ----------
        description : dict
            Configuration from parsing previous files.
        section : string, optional
            What configuration section to read it into if the file does not specify.
        '''
        if description == None:
            description = Description()
            
        self.reader = self.parser_for_extension(file_path)
        self.reader.read(file_path, description, section, **kwargs)
        
        return description
    
    
    def read_string(self, data_string, description=None, section=None, header=None):
        '''Parse data needed for a simulation from a string.'''
        raise Exception("Not implemented, use a sub class")
    
    
    def write(self, filename, description):
        """Save the configuration.  
        
        Parameters
        ----------
        filename : string
            Name of the file to write.
        """
        writer = self.parser_for_extension(filename)
        
        writer.write(filename, description)
    
    
    def parser_for_extension(self, filename):
        '''Choose a subclass that can read the format.
        
        Parameters
        ----------
        filename : string
           For the extension.
           
        Returns
        -------
        DescriptionParser
            Appropriate subclass.
        '''
        # Circular imports
        from allensdk.config.model.formats.json_description_parser import JsonDescriptionParser
        from allensdk.config.model.formats.pycfg_description_parser import PycfgDescriptionParser
        
        parser = None
        
        if filename.endswith('.json'):
            parser = JsonDescriptionParser()
        elif filename.endswith('.pycfg'):
            parser = PycfgDescriptionParser()
        else:
            raise Exception('could not determine file format')
            
        return parser


""" A Generalized Linear Integrate and Fire (GLIF) neuron modeling package. 
Use this code to run the GLIF models available in the Allen Cell Types Atlas.
See :doc:`glif_models` for more details.
"""



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.api.api import Api

import os, json, logging, urllib

class CellTypesApi(Api):
    def __init__(self, base_uri=None):
        super(CellTypesApi, self).__init__(base_uri)


    def list_cells(self, require_morphology=False, require_reconstruction=False):
        ''' Query the API for a list of all cells in the Cell Types Database.

        Parameters
        ----------
        require_morphology: boolean
            Only return cells that have morphology images.

        require_reconstruction: boolean
            Only return cells that have morphological reconstructions.

        Returns
        -------
        list
            Meta data for all cells.
        '''
        
        return self.do_rma_query(self.build_list_cells_rma, self.read_list_cells_json, 
                                 require_morphology, require_reconstruction)


    def build_list_cells_rma(self, require_morphology, require_reconstruction, fmt='json'):
        ''' Build the RMA URL that will fetch meta data for all cells in the cell types database '''

        criteria_associations = "[is_cell_specimen$eq'true'],products[name$eq'Mouse Cell Types']"

        if require_morphology:
            criteria_associations += ",data_sets"

        if require_reconstruction:
            criteria_associations += ",neuron_reconstructions"

        include_associations = 'structure,donor(transgenic_lines),specimen_tags,cell_soma_locations'

        url = ''.join([self.rma_endpoint, 
                       '/query.',
                       fmt,
                       '?q=',
                       'model::Specimen,',
                       'rma::criteria,',
                       criteria_associations,
                       ',rma::include,',
                       include_associations,
                       ',rma::options[num_rows$eqall]'])

        return url


    def read_list_cells_json(self, parsed_json):
        ''' Return the list of cells from the parsed query. '''
        return parsed_json['msg']


    def save_ephys_data(self, specimen_id, file_name):
        ''' Find the electrophysiology data file for a specimen and save it. '''

        
        file_url = self.do_rma_query(self.build_ephys_data_rma, self.read_ephys_data_json, specimen_id)

        if file_url is None:
            raise Exception("Could not find electrophysiology file for specimen %d " % specimen_id)

        self.retrieve_file_over_http(self.api_url + file_url, file_name)
    

    def build_ephys_data_rma(self, specimen_id, fmt='json'):
        ''' Build the URL for the save_ephys_data command. '''
        criteria = '[id$eq%d],ephys_result(well_known_files(well_known_file_type[name$eqNWB]))' % specimen_id
        includes = 'ephys_result(well_known_files(well_known_file_type))'

        url = ''.join([self.rma_endpoint, 
                       '/query.',
                       fmt,
                       '?q=',
                       'model::Specimen,',
                       'rma::criteria,',
                       criteria,
                       ',rma::include,',
                       includes,
                       ',rma::options[num_rows$eqall]'])

        return url


    def read_ephys_data_json(self, parsed_json):
        ''' Extract the download link from the save_ephys_data RMA. '''

        try:
            return parsed_json['msg'][0]['ephys_result']['well_known_files'][0]['download_link']
        except Exception, e:
            logging.error("Error finding ephys file: %s" % e.message)
            return None


    def save_reconstruction(self, specimen_id, file_name):
        ''' Find the reconstruction file for a specimen and save it. '''

        file_url = self.do_rma_query(self.build_reconstruction_rma, self.read_reconstruction_json, specimen_id)

        if file_url is None:
            raise Exception("Could not find reconstruction file for specimen %d " % specimen_id)

        self.retrieve_file_over_http(self.api_url + file_url, file_name)


    def build_reconstruction_rma(self, specimen_id, fmt='json'):
        ''' Build the URL for the save_reconstruction command. '''
        criteria = '[id$eq%d],neuron_reconstructions(well_known_files)' % specimen_id
        includes = 'neuron_reconstructions(well_known_files)'

        url = ''.join([self.rma_endpoint, 
                       '/query.',
                       fmt,
                       '?q=',
                       'model::Specimen,',
                       'rma::criteria,',
                       criteria,
                       ',rma::include,',
                       includes,
                       ',rma::options[num_rows$eqall]'])

        return url


    def read_reconstruction_json(self, parsed_json):
        ''' Extract the download link from the save_ephys_file RMA. '''

        try:
            return parsed_json['msg'][0]['neuron_reconstructions'][0]['well_known_files'][0]['download_link']
        except Exception, e:
            logging.error("Error finding reconstruction file: %s" % e.message)
            return None
    



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import json

def read(file_name):
    """ Shortcut reading JSON from a file. """
    with open(file_name, 'rb') as f:
        return json.loads(f.read())

def write(file_name, obj):
    """ Shortcut for writing JSON to a file.  This also takes care of serializing numpy and data types. """
    with open(file_name, 'wb') as f:
        f.write(json.dumps(obj, indent=2, default=handler))

def handler(obj):
    """ Used by write_json convert a few non-standard types to things that the json package can handle. """
    if hasattr(obj, 'to_dict'):
        return obj.to_dict()
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, np.float64) or isinstance(obj, np.float32):
        return float(obj)
    elif hasattr(obj, 'isoformat'):
        return obj.isoformat()
    else:
        raise TypeError, 'Object of type %s with value of %s is not JSON serializable' % (type(obj), repr(obj))



import numpy

class DatUtilities(object):
    
    @classmethod
    def save_voltage(cls, output_path, v, t):
        '''Save a single voltage output result into a simple text format.
        
        The output file is one t v pair per line.
        
        Parameters
        ----------
        output_path : string
            file name for output
        v : numpy array
            voltage
        t : numpy array
            time
        '''
        data = numpy.transpose(numpy.vstack((t, v)))
        with open (output_path, "w") as f:
            numpy.savetxt(f, data)


# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import urllib2
from json import load
import logging


class Api(object):
    _log = logging.getLogger(__name__)
    default_api_url = 'http://api.brain-map.org'
    
    def __init__(self, api_base_url_string=None):
        if api_base_url_string==None:
            api_base_url_string=Api.default_api_url
            
        self.set_api_urls(api_base_url_string)
        self.default_working_directory = None
    
    
    def set_api_urls(self, api_base_url_string):
        '''Set the internal RMA and well known file download endpoint urls
        based on a api server endpoint.
        
        Parameters
        ----------
        api_base_url_string : string
            url of the api to point to
        '''
        self.api_url = api_base_url_string
        self.well_known_file_endpoint = api_base_url_string + '/api/v2/well_known_file_download/'
        self.rma_endpoint = api_base_url_string + '/api/v2/data'  
    
    
    def set_default_working_directory(self, working_directory):
        '''Set the working directory where files will be saved.
        
        Parameters
        ----------
        working_directory : string
             the absolute path string of the working directory.
        '''
        self.default_working_directory = working_directory
    
    
    def do_rma_query(self, rma_builder_fn, json_traversal_fn, *args, **kwargs):
        '''Bundle an RMA query url construction function
        with a corresponding response json traversal function.
        
        Parameters
        ----------
        rma_builder_fn : function
            A function that takes parameters and returns an rma url.
        json_traversal_fn : function
            A function that takes a json-parsed python data structure and returns data from it.
        args : arguments
            Arguments to be passed to the rma builder function.
        kwargs : keyword arguments
            Keyword arguments to be passed to the rma builder function.
        
        Returns
        -------
        any type
            The data extracted from the json response.
        
        Examples
        --------
        `A simple Api subclass example
        <data_api_client.html#creating-new-api-query-classes>`_.
        '''
        rma_url = rma_builder_fn(*args, **kwargs) 

        quoted_rma_url = urllib2.quote(rma_url,';/?:@&=+$,')
                           
        json_parsed_data = self.retrieve_parsed_json_over_http(quoted_rma_url)
        
        return json_traversal_fn(json_parsed_data)
    
    
    def load_api_schema(self):
        '''Download the RMA schema from the current RMA endpoint
        
        Returns
        -------
        dict
            the parsed json schema message
        
        Notes
        -----
        This information and other 
        `Allen Brain Atlas Data Portal Data Model <http://help.brain-map.org/display/api/Data+Model>`_
        documentation is also available as a
        `Class Hierarchy <http://api.brain-map.org/class_hierarchy>`_
        and `Class List <http://api.brain-map.org/class_hierarchy>`_.
        
        '''
        schema_url = self.rma_endpoint + '/enumerate.json'
        json_parsed_schema_data = self.retrieve_parsed_json_over_http(schema_url)
        
        return json_parsed_schema_data
    
    
    def construct_well_known_file_download_url(self, well_known_file_id):
        '''Join data api endpoint and id.
        
        Parameters
        ----------
        well_known_file_id : integer or string representing an integer
            well known file id
        
        Returns
        -------
        string
            the well-known-file download url for the current api api server
        
        See Also
        --------
        retrieve_file_over_http: Can be used to retrieve the file from the url.
        '''
        return self.well_known_file_endpoint + str(well_known_file_id)
    
    
    def retrieve_file_over_http(self, url, file_path):
        '''Get a file from the data api and save it.
        
        Parameters
        ----------
        url : string
            Url[1]_ from which to get the file.
        file_path : string
            Absolute path including the file name to save.
        
        See Also
        --------
        construct_well_known_file_download_url: Can be used to construct the url.
        
        References
        ----------
        .. [1] Allen Brain Atlas Data Portal: `Downloading a WellKnownFile <http://help.brain-map.org/display/api/Downloading+a+WellKnownFile>`_.
        '''
        try:
            with open(file_path, 'wb') as f:
                response = urllib2.urlopen(url)
                f.write(response.read())
        except urllib2.HTTPError:
            self._log.error("Couldn't retrieve file from %s" % url)
            raise
    
    
    def retrieve_parsed_json_over_http(self, rma_url):
        '''Get the document and put it in a Python data structure
        
        Parameters
        ----------
        rma_url : string
            Full RMA query url.
        
        Returns
        -------
        dict
            Result document as parsed by the JSON library.
        '''
        response = urllib2.urlopen(rma_url)
        json_parsed_data = load(response)
        
        return json_parsed_data



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import re, logging
from pkg_resources import resource_filename #@UnresolvedImport
from allensdk.config.app.application_config import ApplicationConfig
from allensdk.config.model.description_parser import DescriptionParser
from allensdk.config.model.description import Description


class Config(ApplicationConfig):
    _log = logging.getLogger(__name__)
    
    _DEFAULT_LOG_CONFIG = resource_filename(__name__, 'logging.conf')
    
    #: A structure that defines the available configuration parameters.
    #: The default value and help strings may be seen by viewing the source.
    _DEFAULTS = { 
        'workdir': { 'default': 'workdir',
                     'help': 'writable directory where intermediate and output files are written.' },
        'data_dir': { 'default': '',
                      'help': 'writable directory where intermediate and output files are written.' },
        'model_file': { 'default': 'config.json',
                        'help': 'file where the model parameters are set.' },
        'main': { 'default': 'simulation#run',
                  'help' : 'module#function that runs the actual simulation' }
    }
    
    
    def __init__(self):
        super(Config, self).__init__(Config._DEFAULTS, 
                                     name='biophys', 
                                     halp='tools for biophysically detailed modeling at the Allen Institute.',
                                     default_log_config=Config._DEFAULT_LOG_CONFIG)
    
    
    def load(self, config_path,
             disable_existing_logs=False):
        '''Parse the application configuration then immediately load
        the model configuration files.
        
        Parameters
        ----------
        disable_existing_logs : boolean, optional
            If false (default) leave existing logs after configuration.
        '''
        super(Config, self).load([config_path], disable_existing_logs)
        description = self.read_model_description()
        
        return description
        
        
    def read_model_description(self):
        '''parse the model_file field of the application configuration
        and read the files.
        
        The model_file field of the application configuration is
        first split at commas, since it may list more than one file.
        
        The files may be uris of the form :samp:`file:filename?section=name`,
        in which case a bare configuration object is read from filename
        into the configuration section with key 'name'.
        
        A simple filename without a section option
        is treated as a standard multi-section configuration file.
        
        Returns
        -------
        description : Description
            Configuration object.
        '''
        reader = DescriptionParser()
        description = Description()
        
        Config._log.info("model file: %s" % self.model_file)
        
        # TODO: make space aware w/ regex
        for model_file in self.model_file.split(','):
            if not model_file.startswith("file:"):
                model_file = 'file:' + model_file
            
            file_regex=re.compile(r"^file:([^?]*)(\?(.*)?)?")
            m = file_regex.match(model_file)
            model_file = m.group(1)
            file_url_params = {}
            if m.group(3):
                file_url_params.update(((x[0], x[1]) \
                    for x in (y.split('=') \
                        for y in m.group(3).split('&'))))
            if 'section' in file_url_params:
                section = file_url_params['section']
            else:
                section = None
            Config._log.info("reading model file %s" % (model_file))
            reader.read(model_file, description, section)
        
        return description



# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import logging

from allensdk.config.model.manifest import Manifest


class Description(object):
    _log = logging.getLogger(__name__)
    
    def __init__(self):
        self.data = {}
        self.reserved_data = []
        self.manifest = Manifest()
    
    
    def update_data(self, data, section=None):
        '''Merge configuration data possibly from multiple files.
        
        Parameters
        ----------
        data : dict
            Configuration structure to add.
        section : string, optional
            What configuration section to read it into if the file does not specify.
        '''
        if section == None:
            for (section, entries) in data.items():
                if not section in self.data:
                    self.data[section] = entries
                else:
                    self.data[section].extend(entries)
        else:
            if not section in self.data:
                self.data[section] = []
            
            self.data[section].append(data)
    
    
    def is_empty(self):
        '''Check if anything is in the object.
        
        Returns
        -------
        boolean
            true if self.data is missing or empty
        '''
        if self.data:
            return False
        
        return True
    
    
    def unpack(self, data, section=None):
        '''Read the manifest and other stand-alone configuration structure,
        or insert a configuration object into a section of an existing configuration.
        
        Parameters
        ----------
        data : dict
            A configuration object including top level sections,
            or an configuration object to be placed within a section.
        section : string, optional.
            If this is present, place data within an existing section array.
        '''
        if section == None:
            self.unpack_manifest(data)
            self.update_data(data)
        else:
            self.update_data(data, section)
    
    
    def unpack_manifest(self, data):
        '''Pull the manifest configuration section into a separate place.
        
        Parameters
        ----------
        data : dict
            A configuration structure that still has a manifest section.
        '''
        data_manifest = data.pop("manifest", {})
        reserved_data = { "manifest": data_manifest }
        self.reserved_data.append(reserved_data)
        self.manifest.load_config(data_manifest)
    
    
    def fix_unary_sections(self, section_names=None):
        ''' Wrap section contents that don't have the proper
        array surrounding them in an array.
        
        Parameters
        ----------
        section_names : list of strings, optional
            Keys of sections that might not be in array form.
        '''
        if section_names is None:
            section_names = []
        
        for section in section_names:
            if section in self.data:
                if type(self.data[section]) is dict:
                    self.data[section] = [ self.data[section] ];
                    Description._log.warn("wrapped description section %s in an array." % (section))




# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from pkg_resources import resource_filename #@UnresolvedImport
import os
import subprocess as sp
import logging
from allensdk.model.biophys_sim.config import Config


def choose_bps_command(command='bps_simple', conf_file=None):
    log = logging.getLogger('allensdk.model.biophys_sim.bps_command')
    
    log.info("bps command: %s" % (command))
    
    if conf_file:
        conf_file = os.path.abspath(conf_file)
    
    if command == 'help':
        print Config().argparser.parse_args(['--help'])
    elif command == 'nrnivmodl':
        sp.call(['nrnivmodl', 'modfiles']) # TODO: alternate location in manifest?
    elif command == 'run_simple':
        app_config = Config()
        description = app_config.load(conf_file)
        sys.path.insert(1, description.manifest.get_path('CODE_DIR'))
        (module_name, function_name) = description.data['runs'][0]['main'].split('#')
        run_module(description, module_name, function_name)
    else:
        raise Exception("unknown command %s" %(command))


def run_module(description, module_name, function_name):
    m = __import__(module_name, fromlist=[function_name])
    func = getattr(m, function_name)
    
    func(description)


# this module is designed to be called from the bps script,
# which may use nrniv which does not pass in command line arguments
# So the configuration file path must be set in an environment variable.
if __name__ == '__main__':
    import sys
    conf_file = None
    argv = sys.argv
    
    if len(argv) > 1:
        if argv[0] == 'nrniv':
            command = 'run_simple'
        else:
            command = argv[1]
    else:        
        command = 'run_simple'
    
    if len(argv) > 2 and (argv[-1].endswith('.conf') or
                          argv[-1].endswith('.json')):
        conf_file = argv[-1]
    else:    
        try:
            conf_file = os.environ['CONF_FILE']
        except:
            pass
    
    choose_bps_command(command, conf_file)


# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import logging

import numpy as np
import json
import copy

import allensdk.core.json_utilities as json_utilities

from glif_neuron_methods import GlifNeuronMethod, METHOD_LIBRARY

class GlifBadResetException( Exception ):
    """ Exception raised when voltage is still above threshold after a reset rule is applied. """
    def __init__(self, message, dv):
        super(Exception, self).__init__(message)
        self.dv = dv
            

class GlifNeuron( object ):    
    """ Implements the current-based Mihalas Neiber GLIF neuron.  Simulations model the voltage, 
    threshold, and afterspike currents of a neuron given an input stimulus.  A set of modular dynamics
    rules are applied until voltage crosses threshold, at which point a set of modular reset rules are 
    applied. See glif_neuron_methods.py for a list of what options there are for voltage, threshold, and
    afterspike current dynamics and reset rules.

    Parameters
    ----------
    El : float 
            resting potential 
    dt : float
        duration between time steps
    tau : np.ndarray
        time constants of the after-spike currents
    R_input : float
        input resistance
    C : float
        capacitance
    asc_vector : np.ndarray
        afterspike current amplitude vector.  one element per element of tau.
    spike_cut_length : int
        how many time steps to replace with NaNs when a spike occurs.
    th_inf : float
        instantaneous threshold
    coeffs : dict
        dictionary coefficients premultiplied to neuron properties during simulation. used for optimization.
    AScurrent_dynamics_method : dict
        dictionary containing the 'name' of the afterspike current dynamics method to use and a 'params' dictionary parameters to pass to that function.
    voltage_dynamics_method : dict
        dictionary containing the 'name' of the voltage dynamics method to use and a 'params' dictionary parameters to pass to that function.
    threshold_dynamics_method : dict
        dictionary containing the 'name' of the threshold dynamics method to use and a 'params' dictionary parameters to pass to that function.
    AScurrent_reset_method : dict
        dictionary containing the 'name' of the afterspike current dynamics method to use and a 'params' dictionary parameters to pass to that function.
    voltage_reset_method : dict
        dictionary containing the 'name' of the voltage dynamics method to use and a 'params' dictionary parameters to pass to that function.
    threshold_reset_method : dict
        dictionary containing the 'name' of the threshold dynamics method to use and a 'params' dictionary parameters to pass to that function.
    init_voltage : float 
        initial voltage value
    init_threshold : float
        initial spike threshold value
    init_AScurrents : np.ndarray
        initial afterspike current amplitude vector. one element per element of tau.
    """

    TYPE = "GLIF"

    def __init__(self, El, dt, tau, R_input, C, asc_vector, spike_cut_length, th_inf, th_adapt, coeffs,
                 AScurrent_dynamics_method, voltage_dynamics_method, threshold_dynamics_method,
                 AScurrent_reset_method, voltage_reset_method, threshold_reset_method,
                 init_method_data, init_voltage, init_threshold, init_AScurrents, **kwargs): 

        """ Initialize the neuron. """

        self.type = GlifNeuron.TYPE
        self.El = El
        self.dt = dt
        self.tau = np.array(tau)
        
        self.R_input = R_input
        self.C = C

        self.asc_vector = np.array(asc_vector)
        self.spike_cut_length = int(spike_cut_length)
        self.th_inf = th_inf
        self.th_adapt=th_adapt

        self.init_method_data = init_method_data
        self.update_method_data = copy.deepcopy(init_method_data)

        self.init_voltage = init_voltage
        self.init_threshold = init_threshold
        self.init_AScurrents = init_AScurrents

        assert len(tau) == len(asc_vector), Exception("After-spike current vector must have same length as tau (%d vs %d)" % (asc_vector, tau))
        assert len(self.init_AScurrents) == len(self.tau), Exception("init_AScurrents length (%d) must have same length as tau (%d)" % (len(self.init_AScurrents), len(self.tau)))


        # values computed based on inputs
        self.k = 1.0 / self.tau
        self.G = 1.0 / self.R_input

        # Values that can be fit: They scale the input values.  
        # These are allowed to have default values because they are going to get optimized.
        self.coeffs = {
            'th_inf': 1.0,
            'C': 1.0,
            'G': 1.0,
            'b': 1.0,
            'a': 1.0,
            'asc_vector': np.ones(len(self.tau))
        }

        self.coeffs.update(coeffs)
        
        logging.debug('spike cut length: %d' %  self.spike_cut_length)

        # initialize dynamics methods
        self.AScurrent_dynamics_method = self.configure_library_method('AScurrent_dynamics_method', AScurrent_dynamics_method)
        self.voltage_dynamics_method = self.configure_library_method('voltage_dynamics_method', voltage_dynamics_method)
        self.threshold_dynamics_method = self.configure_library_method('threshold_dynamics_method', threshold_dynamics_method)

        # initialize reset methods
        self.AScurrent_reset_method = self.configure_library_method('AScurrent_reset_method', AScurrent_reset_method)
        self.voltage_reset_method = self.configure_library_method('voltage_reset_method', voltage_reset_method)
        self.threshold_reset_method = self.configure_library_method('threshold_reset_method', threshold_reset_method)

    def __str__(self):
        """ Convert an instance of the neuron to a string """
        return json.dumps(self.to_dict(), default=json_utilities.handler, indent=2)

    @property
    def tau_m(self):
        """ Return the cell's membrane tau, which is the product of its input resistance and capacitance. """
        return self.R_input*self.C

    @classmethod
    def from_dict(cls, d):
        """ Initialize a GlifNeuron instance from a dictionary of initialization parameters. """

        return cls(El = d['El'],
                   dt = d['dt'],
                   tau = d['tau'],
                   R_input = d['R_input'],
                   C = d['C'],
                   asc_vector = d['asc_vector'],
                   spike_cut_length = d['spike_cut_length'],
                   th_inf = d['th_inf'],
                   th_adapt=d['th_adapt'],
                   coeffs = d.get('coeffs', {}),
                   AScurrent_dynamics_method = d['AScurrent_dynamics_method'],
                   voltage_dynamics_method = d['voltage_dynamics_method'],
                   threshold_dynamics_method = d['threshold_dynamics_method'],
                   voltage_reset_method = d['voltage_reset_method'],
                   AScurrent_reset_method = d['AScurrent_reset_method'],
                   threshold_reset_method = d['threshold_reset_method'],
                   init_method_data = d.get('init_method_data', {}),
                   init_voltage = d['init_voltage'],
                   init_threshold = d['init_threshold'],
                   init_AScurrents = d['init_AScurrents'])

    def to_dict(self):
        """ Convert the neuron to a serializable dictionary. """

        return {
            'type': self.type,
            'El': self.El,
            'dt': self.dt,
            'tau': self.tau,
            'R_input': self.R_input,
            'C': self.C,
            'asc_vector': self.asc_vector,
            'spike_cut_length': self.spike_cut_length,
            'th_inf': self.th_inf,
            'coeffs': self.coeffs,
            'AScurrent_dynamics_method': self.AScurrent_dynamics_method,
            'voltage_dynamics_method': self.voltage_dynamics_method,
            'threshold_dynamics_method': self.threshold_dynamics_method,
            'AScurrent_reset_method': self.AScurrent_reset_method,
            'voltage_reset_method': self.voltage_reset_method,
            'threshold_reset_method': self.threshold_reset_method,
            'init_method_data': self.init_method_data,
            'update_method_data': self.update_method_data,
            'init_voltage': self.init_voltage,
            'init_threshold': self.init_threshold,
            'init_AScurrents': self.init_AScurrents
        }

    @staticmethod
    def configure_method(method_name, method, method_params):
        """ Create a GlifNeuronMethod instance given a name, a function, and function parameters. 
        This is just a shortcut to the GlifNeuronMethod constructor.

        Parameters
        ----------
        method_name : string
            name for refering to this method later
        method : function
            a python function 
        method_parameters : dict
           function arguments whose values should be fixed

        Returns
        -------
        GlifNeuronMethod
            a GlifNeuronMethod instance
        """ 

        return GlifNeuronMethod(method_name, method, method_params)

    @staticmethod
    def configure_library_method(method_type, params):
        """ Create a GlifNeuronMethod instance out of a library of functions organized by type name. 
        This refers to the METHOD_LIBRARY in glif_neuron_methods.py, which lays out the available functions 
        that can be used for dynamics and reset rules.

        Parameters
        ----------
        method_type : string
            the name of a function category (e.g. 'AScurrent_dynamics_method' for the afterspike current dynamics methods)
        params : dict
            a dictionary with two members. 'name': the string name of function you want, and 'params': parameters you want to pass to that function

        Returns
        -------
        GlifNeuronMethod
            a GlifNeuronMethod instance
        """
        method_options = METHOD_LIBRARY.get(method_type, None)

        assert method_options is not None, Exception("Unknown method type (%s)" % method_type)
        
        method_name = params.get('name', None)
        method_params = params.get('params', None)
        
        assert method_name is not None, Exception("Method configuration for %s has no 'name'" % (method_type))
        assert method_params is not None, Exception("Method configuration for %s has no 'params'" % (method_params))
        
        method = method_options.get(method_name, None)
        
        assert method is not None, Exception("unknown method name %s of type %s" % (method_name, method_type))
        
        return GlifNeuron.configure_method(method_name, method, method_params)


    def reset_method_data(self):
        """ Reinitialize self.update_method_data to its initial state (self.init_method_data). """
        self.update_method_data = copy.deepcopy(self.init_method_data)


    def dynamics(self, voltage_t0, threshold_t0, AScurrents_t0, inj, time_step, spike_time_steps):    
        """ Update the voltage, threshold, and afterspike currents of the neuron for a single time step.

        Parameters
        ----------
        voltage_t0 : float
            the current voltage of the neuron
        threshold_t0 : float
            the current spike threshold level of the neuron
        AScurrents_t0 : np.ndarray
            the current state of the afterspike currents in the neuron
        inj : float
            the current value of the current injection into the neuron
        time_step : int
            the current time step of the neuron simulation
        spike_time_steps : list
            a list of all of the time steps of spikes in the neuron

        Returns
        -------
        tuple
            voltage_t1 (voltage at next time step), threshold_t1 (threshold at next time step), AScurrents_t1 (afterspike currents at next time step)
        """

        AScurrents_t1 = self.AScurrent_dynamics_method(self, AScurrents_t0, time_step, spike_time_steps)
        voltage_t1 = self.voltage_dynamics_method(self, voltage_t0, AScurrents_t0, inj)
        threshold_t1 = self.threshold_dynamics_method(self, threshold_t0, voltage_t0)

        return voltage_t1, threshold_t1, AScurrents_t1
    
        
    def reset(self, voltage_t0, threshold_t0, AScurrents_t0):
        """ Apply reset rules to the neuron's voltage, threshold, and afterspike currents assuming a spike has occurred (voltage is above threshold). 

        Parameters
        ----------
        voltage_t0 : float
           the current voltage of the neuron
        threshold_t0 : float
            the current spike threshold level of the neuron
        AScurrents_t0 : np.ndarray
            the current state of the afterspike currents in the neuron

        Returns
        -------
        tuple
            voltage_t1 (voltage at next time step), threshold_t1 (threshold at next time step), AScurrents_t1 (afterspike currents at next time step)
        """
        
        AScurrents_t1 = self.AScurrent_reset_method(self, AScurrents_t0)  
        voltage_t1 = self.voltage_reset_method(self, voltage_t0)
        threshold_t1 = self.threshold_reset_method(self, threshold_t0, voltage_t1)

        if voltage_t1 > threshold_t1:
            raise GlifBadResetException("Voltage reset above threshold: voltage_t1 (%f) threshold_t1 (%f), voltage_t0 (%f) threshold_t0 (%f) AScurrents_t0 (%s)" % ( voltage_t1, threshold_t1, voltage_t0, threshold_t0, repr(AScurrents_t0)), voltage_t1 - threshold_t1)

        return voltage_t1, threshold_t1, AScurrents_t1
    
    def run(self, stim):
        """ Run neuron simulation over a given stimulus. This steps through the stimulus applying dynamics equations.
        After each step it checks if voltage is above threshold.  If so, self.spike_cut_length NaNs are inserted 
        into the output voltages, reset rules are applied to the voltage, threshold, and afterspike currents, and the 
        simulation resumes.

        Parameters
        ----------
        stim : np.ndarray
            vector of scalar current values

        Returns
        -------
        dict
            a dictionary containing: 
                'voltage': simulated voltage values, 
                'threshold': threshold values during the simulation,
                'AScurrents': afterspike current values during the simulation, 
                'grid_spike_times': spike times (in uits of self.dt) aligned to simulation time steps, 
                'interpolated_spike_times': spike times (in units of self.dt) linearly interpolated between time steps, 
                'spike_time_steps': the indices of grid spike times, 
                'interpolated_spike_voltage': voltage of the simulation at interpolated spike times, 
                'interpolated_spike_threshold': threshold of the simulation at interpolated spike times
        """

        # initialize the voltage, threshold, and afterspike current values
        voltage_t0 = self.init_voltage
        threshold_t0 = self.init_threshold
        AScurrents_t0 = self.init_AScurrents

        self.reset_method_data() #get rid of lingering method data

        num_time_steps = len(stim) 
        num_AScurrents = len(AScurrents_t0)
        
        # pre-allocate the output voltages, thresholds, and after-spike currents
        voltage_out=np.empty(num_time_steps)
        voltage_out[:]=np.nan
        threshold_out=np.empty(num_time_steps)
        threshold_out[:]=np.nan
        AScurrents_out=np.empty(shape=(num_time_steps, num_AScurrents))
        AScurrents_out[:]=np.nan        

        # array that will hold spike indices
        spike_time_steps = []
        grid_spike_times = []
        interpolated_spike_times = []
        interpolated_spike_voltage = []
        interpolated_spike_threshold = []

        time_step = 0
        while time_step < num_time_steps:
            if time_step % 10000 == 0:
                logging.info("time step %d / %d" % (time_step,  num_time_steps))

            # compute voltage, threshold, and ascurrents at current time step
            (voltage_t1, threshold_t1, AScurrents_t1) = self.dynamics(voltage_t0, threshold_t0, AScurrents_t0, stim[time_step], time_step, spike_time_steps) 

            #if the voltage is bigger than the threshold record the spike and reset the values
            if voltage_t1 > threshold_t1: 

                # spike_time_steps are stimulus indices when voltage surpassed threshold
                spike_time_steps.append(time_step)
                grid_spike_times.append(time_step * self.dt) 

                # compute higher fidelity spike time/voltage/threshold by linearly interpolating 
                interpolated_spike_times.append(interpolate_spike_time(self.dt, time_step, threshold_t0, threshold_t1, voltage_t0, voltage_t1))

                interpolated_spike_time_offset = interpolated_spike_times[-1] - (time_step - 1) * self.dt
                interpolated_spike_voltage.append(interpolate_spike_value(self.dt, interpolated_spike_time_offset, voltage_t0, voltage_t1))
                interpolated_spike_threshold.append(interpolate_spike_value(self.dt, interpolated_spike_time_offset, threshold_t0, threshold_t1))
            
                # reset voltage, threshold, and afterspike currents
                (voltage_t0, threshold_t0, AScurrents_t0) = self.reset(voltage_t1, threshold_t1, AScurrents_t1) 

                # if we are not integrating during the spike (which includes right now), insert nans then jump ahead
                if self.spike_cut_length > 0:
                    n = self.spike_cut_length
                    voltage_out[time_step:time_step+n] = np.nan
                    threshold_out[time_step:time_step+n] = np.nan
                    AScurrents_out[time_step:time_step+n,:] = np.nan

                    time_step += self.spike_cut_length
                else:
                    # we are integrating during the spike, so store the reset values
                    voltage_out[time_step] = voltage_t0 
                    threshold_out[time_step] = threshold_t0
                    AScurrents_out[time_step,:] = AScurrents_t0

                    time_step += 1
            else:
                # there was no spike, store the next voltages
                voltage_out[time_step] = voltage_t1 
                threshold_out[time_step] = threshold_t1
                AScurrents_out[time_step,:] = AScurrents_t1

                voltage_t0 = voltage_t1
                threshold_t0 = threshold_t1
                AScurrents_t0 = AScurrents_t1
                
                time_step += 1

        return {
            'voltage': voltage_out, 
            'threshold': threshold_out, 
            'AScurrents': AScurrents_out,
            'grid_spike_times': np.array(grid_spike_times), 
            'interpolated_spike_times': np.array(interpolated_spike_times), 
            'spike_time_steps': np.array(spike_time_steps), 
            'interpolated_spike_voltage': np.array(interpolated_spike_voltage), 
            'interpolated_spike_threshold': np.array(interpolated_spike_threshold)
            }


def interpolate_spike_time(dt, time_step, threshold_t0, threshold_t1, voltage_t0, voltage_t1):
    """ Given two voltage and threshold values, the dt between them and the initial time step, interpolate
    a spike time within the dt interval by intersecting the two lines. """
    return time_step*dt + line_crossing_x(dt, voltage_t0, voltage_t1, threshold_t0, threshold_t1)


def interpolate_spike_value(dt, interpolated_spike_time_offset, v0, v1):
    """ Take a value at two adjacent time steps and linearly interpolate what the value would be
    at an offset between the two time steps. """
    return v0 + (v1 - v0) * interpolated_spike_time_offset / dt


def line_crossing_x(dx, a0, a1, b0, b1):
    """ Find the x value of the intersection of two lines. """
    assert type(a0) != int and type(a1) != int and type(b0) != int and type(b1) != int, Exception("Do not pass integers into this function!")
    return dx * (b0 - a0) / ( (a1 - a0) - (b1 - b0) )
        

def line_crossing_y(dx, a0, a1, b0, b1):
    """ Find the y value of the intersection of two lines. """
    return b0 + (b1 - b0) * (b0 - a0) / ((a1 - a0) - (b1 - b0))




# Copyright 2014 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import re, logging
from json import loads
from __builtin__ import classmethod


class JsonUtil(object):
    _log = logging.getLogger(__name__)
    _oneline_comment = re.compile(r"\/\/.*$",
                                  re.MULTILINE)
    _multiline_comment_start = re.compile(r"\/\*",
                                          re.MULTILINE |
                                          re.DOTALL)
    _multiline_comment_end = re.compile(r"\*\/",
                                        re.MULTILINE |
                                        re.DOTALL)
    _blank_line = re.compile(r"\n?^\s*$", re.MULTILINE)
    _carriage_return = re.compile(r"\r$", re.MULTILINE)
    
    
    @classmethod
    def read_json_file(cls, json_path):
        '''Transform a JSON contained in a file into an equivalent
        nested python dict.
        
        Parameters
        ----------
        json_path : string
            The input.
        
        Returns
        -------
        dict
            Python version of the input
        
        Note: if the input is a bare array or literal, for example,
        the output will be of the corresponding type.
        '''
        with open(json_path, 'rb') as f:
            json_string = f.read()
        
        return cls.read_json_string(json_string)
    
    
    @classmethod
    def read_json_string(cls, json_string):
        '''Transform a JSON string into an equivalent nested python dict.
        
        Parameters
        ----------
        json_string : string
            The input.
        
        Returns
        -------
        dict
            Python version of the input
        
        Note: if the input is a bare array or literal, for example,
        the output will be of the corresponding type.
        '''
        return loads(cls.remove_comments(json_string))
    
    
    @classmethod
    def remove_comments(cls, json_string):
        '''Strip single and multiline javascript-style comments.
        
        Parameters
        ----------
        json : string
            Json string with javascript-style comments.
            
        Returns
        -------
        string
            Copy of the input with comments removed.
            
        Note: A JSON decoder MAY accept and ignore comments.
        '''
        json_string = JsonUtil._oneline_comment.sub('', json_string)
        json_string = JsonUtil._carriage_return.sub('', json_string)
        json_string = JsonUtil.remove_multiline_comments(json_string)
        json_string = JsonUtil._blank_line.sub('', json_string)
        
        return json_string
    
    
    @classmethod
    def remove_multiline_comments(cls, json_string):
        '''Rebuild input without substrings matching /*...*/.
        
        Parameters
        ----------
        json_string : string
            may or may not contain multiline comments.
        
        Returns
        -------
        string
            Copy of the input without the comments.
        '''
        new_json = []
        start_iter = JsonUtil._multiline_comment_start.finditer(json_string)
        json_slice_start = 0
        
        for comment_start in start_iter:
            json_slice_end = comment_start.start()
            new_json.append(json_string[json_slice_start:json_slice_end])
            search_start = comment_start.end()
            comment_end = JsonUtil._multiline_comment_end.search(json_string[search_start:])
            if comment_end == None:
                break
            else:
                json_slice_start = search_start + comment_end.end()
        new_json.append(json_string[json_slice_start:])
        
        return ''.join(new_json)



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import re
import csv
import copy

# Default columns to be read and written by this module 
DEFAULT_COLUMNS = [ 'id', 'type', 'x', 'y', 'z', 'radius', 'parent' ]

# Default columns to convert to numeric types automatically 
DEFAULT_NUMERIC_COLUMNS = [ 'type', 'x', 'y', 'z', 'radius' ]

def read_swc(file_name, columns=None, numeric_columns=None):
    """  Read in an SWC file and return a Morphology object.
    SWC are basically CSV files, but they often don't have headers. 
    You can pass those in explicitly and also indicate which
    columns are numeric.

    Parameters
    ----------
    file_name: string
        SWC file name.

    columns: list of strings
        names of the columns in this file (default: DEFAULT_COLUMNS)

    numeric_columns: list of strings
        names of the numeric columns in this file (default: DEFAULT_NUMERIC_COLUMNS)

    Returns
    -------
    Morphology
        A Morphology instance.
    """

    if columns is None:
        columns = DEFAULT_COLUMNS
        
    if numeric_columns is None:
        numeric_columns = DEFAULT_NUMERIC_COLUMNS 

    with open(file_name, "rb") as f:
        # skip comment rows, strip off extra whitespace
        return read_rows(f, columns, numeric_columns)


def read_rows(rows, columns, numeric_columns):
    """ Parse a list of string SWC rows.  Lines that start with '#'
    are ignored.  Numeric types are properly converted.

    Parameters
    ----------
    rows: list of strings
        usually the rows of an SWC file

    columns: list of strings
        names of the columns in this file

    numeric_columns: list of strings
        names of the numeric columns in this file

    Returns
    -------
    Morphology
        A Morphology instance.
    """

    rows = [ r.strip() for r in rows if len(r) > 0 and r.strip()[0] != '#' ]
    
    reader = csv.DictReader(rows, fieldnames=columns, delimiter=' ', skipinitialspace=True, restkey='other')
    
    compartment_list = []

    # convert numeric columns 
    for compartment in reader:
        for nh in numeric_columns:
            compartment[nh] = str_to_num(compartment[nh])
                
        compartment_list.append(compartment)

    return Morphology(compartment_list=compartment_list)    
    

def read_string(s, columns=None, numeric_columns=None):
    """ Parse a list of string SWC rows.  Lines that start with '#'
    are ignored.  Numeric types are properly converted.

    Parameters
    ----------
    s: string
        the contents of an SWC file as a string

    columns: list of strings
        names of the columns in this file (default: DEFAULT_COLUMNS)

    numeric_columns: list of strings
        names of the numeric columns in this file (default: DEFAULT_NUMERIC_COLUMNS)

    Returns
    -------
    Morphology
        A Morphology instance.
    
    """
    if columns is None:
        columns = DEFAULT_COLUMNS
        
    if numeric_columns is None:
        numeric_columns = DEFAULT_NUMERIC_COLUMNS 

    rows = s.split('\n')

    return read_rows(rows, columns, numeric_columns)


class Morphology( object ):
    """ Keep track of the list of compartments in a morphology and provide 
    a few helper methods (index by id, sparsify, root, etc).  During initialization
    the compartments are assigned a 'children' property that is a list of
    pointers to child compartments.
    """

    SOMA = 1

    def __init__(self, compartment_list=None, compartment_index=None):
        """ Try to initialize from a list of compartments first, then from
        a dictionary indexed by compartment id if that fails, and finally just
        leave everything empty.

        Parameters
        ----------
        compartment_list: list 
            list of compartment dictionaries
            
        compartment_index: dict
            dictionary of compartments indexed by id
        """

        self._compartment_list = []
        self._compartment_index = {}
        
        # first try the compartment list, then try the compartment index
        if compartment_list:
            self.compartment_list = compartment_list
        elif compartment_index:
            self.compartment_index = compartment_index


    @property 
    def compartment_list(self):
        """ Return the compartment list.  This is a property to ensure that the 
        compartment list and compartment index are in sync. """
        return self._compartment_list



    @compartment_list.setter
    def compartment_list(self, compartment_list):
        """ Update the compartment list.  Update the compartment index. """
        self._compartment_list = compartment_list
        self._compartment_index = { c['id']: c for c in compartment_list }
        self.update_children()


    @property
    def compartment_index(self):
        """ Return the compartment index.  This is a property to ensure that the
        compartment list and compartment index are in sync. """
        return self._compartment_index


    @compartment_index.setter
    def compartment_index(self, compartment_index):
        """ Update the compartment index.  Update the compartment list. """
        self._compartment_index = compartment_index
        self._compartment_list = compartment_index.values()
        self.update_children()


    @property
    def root(self):
        """ Search through the compartment index for the root compartment """
        for cid,c in self.compartment_index.iteritems():
            if c['parent'] == '-1':
                return c
        return None


    def compartment_index_by_type(self, compartment_type):
        """ Return an dictionary of compartments indexed by id that all have
        a particular compartment type.

        Parameters
        ----------
        compartment_type: int
            Desired compartment type
        """

        return { c['id']: c for c in self._compartment_list if c['type'] == compartment_type }


    def write(self, file_name, columns=None):
        """ Write this morphology out to an SWC file 
      
        Parameters
        ----------
        file_name: string
            desired name of your SWC file

        columns: list
            columns to write to your SWC file (default: DEFAULT_COLUMNS)
        """
        if columns is None:
            columns = DEFAULT_COLUMNS
        
        # convert to array, sort by id
        sorted_compartments = sorted(self.compartment_list, key=lambda x: int(x['id']))

        with open(file_name, 'wb') as f:
            writer = csv.DictWriter(f, delimiter=' ', fieldnames=columns, extrasaction='ignore')
            writer.writerows(sorted_compartments)


    def validate(self):
        """ Make sure that the parents and children are assigned properly. """
        compartments = self._compartment_index

        for cid, c in self.compartment_index.iteritems():
            if c['parent'] != "-1":
                assert c['parent'] in compartments, "bad parent id: %s" % (c['parent'] )
                
            for child in c['children']:
                assert child in compartments, "bad child id: %s" % (child)

        
    def update_children(self):
        """ Fill each compartment's array of children """
        compartments = self._compartment_index

        for i,compartment in compartments.iteritems():
            compartment['children'] = []

        for i,compartment in compartments.iteritems():
            pi = compartment['parent']
    
            if pi not in compartments:
                continue 

            parent = compartments[pi]
            
            parent['children'].append(compartment['id'])


    def sparsify(self, modulo, compress_ids=False):
        """ Return a new Morphology object that has a given number of non-leaf,
        non-root segments removed.  IDs can be reassigned so as to be continuous.

        Parameters
        ----------
        modulo: int
           keep 1 out of every modulo segments.

        compress_ids: boolean
           Reassign ids so that ids are continuous (no missing id numbers).

        Returns
        -------   
        Morphology
            A new morphology instance
        """
        
        compartments = copy.deepcopy(self.compartment_index)
        root = self.root

        keep = {}

        # figure out which compartments to toss
        ct = 0
        for i, c in compartments.iteritems():
            pid = c['parent']
            cid = c['id']
            ctype = c['type']

            # keep the root, soma, junctions, and the first child of the root (for visualization)
            if pid == "-1" or len(c['children']) != 1 or pid == root['id'] or ctype == Morphology.SOMA:
                keep[cid] = True
            else:
                keep[cid] = (ct % modulo) == 0
                
            ct += 1

        # hook children up to their new parents
        for i, c in compartments.iteritems():
            comp_id = c['id']

            if keep[comp_id] is False:
                parent_id = c['parent']
                while keep[parent_id] is False:
                    parent_id = compartments[parent_id]['parent']

                for child_id in c['children']:
                    compartments[child_id]['parent'] = parent_id

        # filter out the orphans
        sparsified_compartments = { k:v for k,v in compartments.iteritems() if keep[k] }

        if compress_ids:
            ids = sorted(sparsified_compartments.keys(), key=lambda x: int(x))
            id_hash = { fid:str(i+1) for i,fid in enumerate(ids) }
            id_hash["-1"] = "-1"

            # build the final compartment index
            out_compartments = {}
            for cid, compartment in sparsified_compartments.iteritems():
                compartment['id'] = id_hash[cid]
                compartment['parent'] = id_hash[compartment['parent']]
                out_compartments[compartment['id']] = compartment

            return Morphology(compartment_index=out_compartments)
        else:
            return Morphology(compartment_index=sparsified_compartments)


def str_to_num(s):
    """ Try to convert a string s into a number """
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s



from setuptools import setup, find_packages
import os
import allensdk

# http://bugs.python.org/issue8876#msg208792
if hasattr(os, 'link'):
    del os.link

def prepend_find_packages(*roots):
    ''' Recursively traverse nested packages under the root directories
    '''
    packages = []
    
    for root in roots:
        packages += [root]
        packages += [root + '.' + s for s in find_packages(root)]
        
    return packages

setup(
    version = allensdk.__version__,
    name = 'allensdk',
    author = 'David Feng',
    author_email = 'davidf@alleninstitute.org',
    packages = prepend_find_packages('allensdk'),
    package_data={'': ['*.conf', '*.cfg', '*.md', '*.json', '*.dat', '*.env', '*.sh', 'bps'] },
    description = 'core libraries for the allensdk.',
    requires = ['h5py',
                'argparse',
                'six'],
    tests_require=['nose>=1.2.1',
                   'coverage>=3.7.1',
                   'mock'],
    setup_requires=['setuptools', 'sphinx', 'numpydoc'],
    url='https://github.com/AllenInstitute/AllenSDK/tree/v%s' % (allensdk.__version__),
    download_url = 'https://github.com/AllenInstitute/AllenSDK/tarball/v%s' % (allensdk.__version__),
    keywords = ['neuroscience', 'bioinformatics', 'scientific'  ],
    scripts=['allensdk/model/biophys_sim/scripts/bps'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
        ])



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.model.biophys_sim.config import Config
from allensdk.model.biophysical_perisomatic.utils import Utils
from allensdk.core.nwb_data_set import NwbDataSet
from shutil import copy
import numpy
from allensdk.core.dat_utilities import DatUtilities


def run(description):
    '''Main function for running a perisomatic biophysical experiment.
    
    Parameters
    ----------
    description : Config
        All information needed to run the experiment.
    '''
    # configure NEURON
    utils = Utils(description)
    h = utils.h
    
    # configure model
    manifest = description.manifest
    morphology_path = description.manifest.get_path('MORPHOLOGY')
    utils.generate_morphology(morphology_path.encode('ascii', 'ignore'))
    utils.load_cell_parameters()
    
    # configure stimulus and recording
    stimulus_path = description.manifest.get_path('stimulus_path')
    run_params = description.data['runs'][0]
    sweeps = run_params['sweeps']
    junction_potential = description.data['fitting'][0]['junction_potential']
    mV = 1.0e-3
    
    stimulus_format = manifest.get_format('stimulus_path')
    output_format = manifest.get_format('output')
    
    if stimulus_format == 'NWB' and output_format == 'NWB':
        prepare_nwb_output(manifest.get_path('stimulus_path'),
                           manifest.get_path('output'))
    
    # run sweeps
    for sweep in sweeps:
        if stimulus_format == 'NWB':
            utils.setup_iclamp(stimulus_path, sweep=sweep)
        elif stimulus_format == 'dat':
            utils.setup_iclamp_dat(stimulus_path)
        
        vec = utils.record_values()
        
        h.finitialize()
        h.run()
        
        # write to an NWB File
        output_data = (numpy.array(vec['v']) - junction_potential) * mV
        output_times = numpy.array(vec['t'])
        
        if output_format == 'NWB':
            output_path = manifest.get_path("output")
            save_nwb(output_path, output_data, sweep)
        elif output_format == 'dat':
            output_path = manifest.get_path("output", sweep)
            DatUtilities.save_voltage(output_path, output_data, output_times)


def prepare_nwb_output(nwb_stimulus_path,
                       nwb_result_path):
    '''Copy the stimulus file, zero out the recorded voltages and spike times.
    
    Parameters
    ----------
    nwb_stimulus_path : string
        NWB file name
    nwb_result_path : string
        NWB file name
    '''
    copy(nwb_stimulus_path, nwb_result_path)
    data_set = NwbDataSet(nwb_result_path)
    data_set.fill_sweep_responses(0.0)
    for sweep in data_set.get_sweep_numbers():
        data_set.set_spike_times(sweep, [])


def save_nwb(output_path, v, sweep):
    '''Save a single voltage output result into an existing sweep in a NWB file.
    This is intended to overwrite a recorded trace with a simulated voltage.
    
    Parameters
    ----------
    output_path : string
        file name of a pre-existing NWB file.
    v : numpy array
        voltage
    sweep : integer
        which entry to overwrite in the file.
    '''
    output = NwbDataSet(output_path)
    output.set_sweep(sweep, None, v)


def load_description(manifest_json_path):
    '''Read configuration file.
    
    Parameters
    ----------
    manifest_json_path : string
        File containing the experiment configuration.
    
    Returns
    -------
    Config
        Object with all information needed to run the experiment.
    '''
    description = Config().load(manifest_json_path)
    
    # fix nonstandard description sections
    fix_sections = ['passive', 'axon_morph,', 'conditions', 'fitting']
    description.fix_unary_sections(fix_sections)
    
    return description


if '__main__' == __name__:
    import sys
    
    description = load_description(sys.argv[-1])
    
    run(description)




# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# Merchantability Or Fitness FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import h5py
import numpy as np

class NwbDataSet(object):
    """ A very simple interface for exracting electrophysiology data
    from an NWB file.
    """

    def __init__(self, file_name):
        """ Initialize the NwbDataSet instance with a file name

        Parameters
        ----------
        file_name: string
           NWB file name
        """
        self.file_name = file_name
    
    def get_sweep(self, sweep_number):
        """ Retrieve the stimulus, response, index_range, and sampling rate
        for a particular sweep.  This method hides the NWB file's distinction 
        between a "Sweep" and an "Experiment".  An experiment is a subset of
        of a sweep that excludes the initial test pulse.  It also excludes 
        any erroneous response data at the end of the sweep (usually for
        ramp sweeps, where recording was terminated mid-stimulus).  

        Some sweeps do not have an experiment, so full data arrays are 
        returned.  Sweeps that have an experiment return full data arrays
        (include the test pulse) with any erroneous data trimmed from the 
        back of the sweep.  

        Parameters
        ----------
        sweep_number: int

        Returns
        -------
        dict
            A dictionary with 'stimulus', 'response', 'index_range', and
            'sampling_rate' elements.  The index range is a 2-tuple where
            the first element indicates the end of the test pulse and the
            second index is the end of valid response data.
        """
        with h5py.File(self.file_name,'r') as f:
            
            swp = f['epochs']['Sweep_%d' % sweep_number]
            
            stimulus = swp['stimulus']['timeseries']['data'].value
            response = swp['response']['timeseries']['data'].value
            
            swp_idx_start = swp['stimulus']['idx_start'].value
            swp_length = swp['stimulus']['count'].value

            swp_idx_stop = swp_idx_start + swp_length - 1
            sweep_index_range = ( swp_idx_start, swp_idx_stop )

            # if the sweep has an experiment, extract the experiment's index range
            try:
                exp = f['epochs']['Experiment_%d' % sweep_number]
                exp_idx_start = exp['stimulus']['idx_start'].value
                exp_length = exp['stimulus']['count'].value
                exp_idx_stop = exp_idx_start + exp_length - 1
                experiment_index_range = ( exp_idx_start, exp_idx_stop )
            except KeyError, _:
                # this sweep has no experiment.  return the index range of the entire sweep.
                experiment_index_range = sweep_index_range
            
            assert sweep_index_range[0] == 0, Exception("index range of the full sweep does not start at 0.")
            
            # only return data up to the end of the experiment -- ignore everything else
            return  {
                'stimulus': stimulus[sweep_index_range[0]:experiment_index_range[1]+1],
                'response': response[sweep_index_range[0]:experiment_index_range[1]+1],
                'index_range': experiment_index_range,
                'sampling_rate': 1.0 * swp['stimulus']['timeseries']['starting_time'].attrs['rate']
            }
    
    
    def set_sweep(self, sweep_number, stimulus, response):
        """ Overwrite the stimulus or response of an NWB file.  
        If the supplied arrays are shorter than stored arrays,
        they are padded with zeros to match the original data 
        size.

        Parameters
        ----------
        sweep_number: int

        stimulus: np.array
           Overwrite the stimulus with this array.  If None, stimulus is unchanged.

        response: np.array
            Overwrite the response with this array.  If None, response is unchanged.
        """

        with h5py.File(self.file_name,'r+') as f:
            swp = f['epochs']['Sweep_%d' % sweep_number]
            
            # this is the length of the entire sweep data, including test pulse and 
            # whatever might be in front of it
            # TODO: remove deprecated 'idx_stop'
            if 'idx_stop' in swp['stimulus']:
                sweep_length = swp['stimulus']['idx_stop'].value + 1
            else:
                sweep_length = swp['stimulus']['count'].value
            
            if stimulus is not None:
                # if the data is shorter than the sweep, pad it with zeros
                missing_data = sweep_length - len(stimulus)
                if missing_data > 0:
                    stimulus = np.append(stimulus, np.zeros(missing_data))
                
                swp['stimulus']['timeseries']['data'][...] = stimulus
            
            if response is not None:
                # if the data is shorter than the sweep, pad it with zeros
                missing_data = sweep_length - len(response)
                if missing_data > 0:
                    response = np.append(response, np.zeros(missing_data))
                
                
                swp['response']['timeseries']['data'][...] = response
    

    def get_spike_times(self, sweep_number):
        """ Return any spike times stored in the NWB file for a sweep.

        Parameters
        ----------
        sweep_number: int

        Returns
        -------
        list
           list of spike times in seconds relative to the start of the sweep
        """
        with h5py.File(self.file_name,'r') as f:
            sweep_name = "Sweep_%d" % sweep_number
            
            try:
                spikes = f["analysis"]["spike_times"][sweep_name]
            except KeyError:
                return []
            
            return spikes.value
    
    
    def set_spike_times(self, sweep_number, spike_times):
        """ Set or overwrite the spikes times for a sweep.
        
        Parameters
        ----------
        sweep_number: int
        
        spike_times: np.array
           array of spike times in seconds
        """
        with h5py.File(self.file_name,'r+') as f:
            # make sure expected directory structure is in place
            if "analysis" not in f.keys():
                f.create_group("analysis")
            
            analysis_dir = f["analysis"]
            if "spike_times" not in analysis_dir.keys():
                analysis_dir.create_group("spike_times")
            
            spike_dir = analysis_dir["spike_times"]
            
            # see if desired dataset already exists
            sweep_name = "Sweep_%d" % sweep_number
            if sweep_name in spike_dir.keys():
                # rewriting data -- delete old dataset
                del spike_dir[sweep_name]
            
            spike_dir.create_dataset(sweep_name, data=spike_times, dtype='f8')


    def get_sweep_numbers(self):
        """ Get all of the sweep numbers in the file, including test sweeps. """
        
        with h5py.File(self.file_name, 'r') as f:
            sweeps = [int(e.split('_')[1]) for e in f['epochs'].keys() if e.startswith('Sweep_')]
            return sweeps


    def get_experiment_sweep_numbers(self):
        """ Get all of the sweep numbers for experiment epochs in the file, not including test sweeps. """
        
        with h5py.File(self.file_name, 'r') as f:
            sweeps = [int(e.split('_')[1]) for e in f['epochs'].keys() if e.startswith('Experiment_')]
            return sweeps
        

    def fill_sweep_responses(self, fill_value=0.0, sweep_numbers=None):
        """ Fill sweep response arrays with a single value.

        Parameters
        ----------
        fill_value: float
            Value used to fill sweep response array
        
        sweep_numbers: list
            List of integer sweep numbers to be filled (default all sweeps)
            
        """

        with h5py.File(self.file_name, 'a') as f:
            if sweep_numbers is None:
                # no sweep numbers given, grab all of them
                epochs = [ k for k in f['epochs'].keys() if k.startswith('Sweep_') ]
            else:
                epochs = [ 'Sweep_%d' % sweep_number for sweep_number in sweep_numbers ]

            for epoch in epochs:
                if epoch in f['epochs']:
                    f['epochs'][epoch]['response']['timeseries']['data'][...] = fill_value



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import logging
from allensdk.model.biophys_sim.neuron.hoc_utils import HocUtils
from allensdk.core.nwb_data_set import NwbDataSet


class Utils(HocUtils):
    '''A helper class for NEURON functionality needed for
    perisomatic biophysical simulations.
    
    Attributes
    ----------
    h : object
        The NEURON hoc object.
    nrn : object
        The NEURON python object.
    neuron : module
        The NEURON module.
    '''
    
    _log = logging.getLogger(__name__)
    
    def __init__(self, description):
        super(Utils, self).__init__(description)
        self.stim = None
        self.stim_curr = None
        self.sampling_rate = None
    
        self.stim_vec_list = []
    
    
    def generate_morphology(self, morph_filename):
        '''Load a swc-format cell morphology file.
        
        Parameters
        ----------
        morph_filename : string
            Path to swc.
        '''
        h = self.h
        
        swc = self.h.Import3d_SWC_read()
        swc.input(morph_filename)
        imprt = self.h.Import3d_GUI(swc, 0)
        
        h("objref this")
        imprt.instantiate(h.this)
        
        h("soma[0] area(0.5)")
        for sec in h.allsec():
            sec.nseg = 1 + 2 * int(sec.L / 40)
            if sec.name()[:4] == "axon":
                h.delete_section(sec=sec)
        h('create axon[2]')
        for sec in h.axon:
            sec.L = 30
            sec.diam = 1
            sec.nseg = 1 + 2 * int(sec.L / 40)
        h.axon[0].connect(h.soma[0], 0.5, 0)
        h.axon[1].connect(h.axon[0], 1, 0)
        
        h.define_shape()
    
    
    def load_cell_parameters(self):
        '''Configure a neuron after the cell morphology has been loaded.'''
        passive = self.description.data['passive'][0]
        genome = self.description.data['genome']
        conditions = self.description.data['conditions'][0]
        h = self.h
        
        h("access soma")
        
        # Set fixed passive properties
        for sec in h.allsec():
            sec.Ra = passive['ra']
            sec.insert('pas')
            for seg in sec:
                seg.pas.e = passive["e_pas"]
        
        for c in passive["cm"]:
            h('forsec "' + c["section"] + '" { cm = %g }' % c["cm"])
        
        # Insert channels and set parameters
        for p in genome:
            if p["section"] == "glob": # global parameter
                h(p["name"] + " = %g " % p["value"])
            else:
                if p["mechanism"] != "":
                    h('forsec "' + p["section"] + '" { insert ' + p["mechanism"] + ' }')
                h('forsec "' + p["section"] + '" { ' + p["name"] + ' = %g }' % p["value"])
        
        # Set reversal potentials
        for erev in conditions['erev']:
            h('forsec "' + erev["section"] + '" { ek = %g }' % erev["ek"])
            h('forsec "' + erev["section"] + '" { ena = %g }' % erev["ena"])
    
    
    def setup_iclamp(self,
                     stimulus_path,
                     sweep=0):
        '''Assign a current waveform as input stimulus.
        
        Parameters
        ----------
        stimulus_path : string
            NWB file name
        '''
        self.stim = self.h.IClamp(self.h.soma[0](0.5))  # TODO: does soma have to be parametrized?
        self.stim.amp = 0
        self.stim.delay = 0
        self.stim.dur = 1e12 # just set to be really big; doesn't need to match the waveform
        
        self.read_stimulus(stimulus_path, sweep=sweep)
        self.h.dt = self.sampling_rate
        stim_vec = self.h.Vector(self.stim_curr)
        stim_vec.play(self.stim._ref_amp, self.sampling_rate)
        
        stimulus_stop_index = len(self.stim_curr) - 1
        self.h.tstop = stimulus_stop_index * self.sampling_rate
        self.stim_vec_list.append(stim_vec)
    
    
    def read_stimulus(self, stimulus_path, sweep=0):
        '''load current values for a specific experiment sweep.
        
        Parameters
        ----------
        stimulus path : string
            NWB file name
        sweep : integer, optional
            sweep index
        '''
        Utils._log.info("reading stimulus path: %s, sweep %s" %
                        (stimulus_path, sweep))
        stimulus_data = NwbDataSet(stimulus_path)
        sweep_data = stimulus_data.get_sweep(sweep)
        self.stim_curr = sweep_data['stimulus'] * 1.0e9 # convert to nA for NEURON
        self.sampling_rate = 1.0e3 / sweep_data['sampling_rate'] # convert from Hz
    
    
    def record_values(self):
        '''Set up output voltage recording.'''
        vec = { "v": self.h.Vector(),
                "t": self.h.Vector() }
    
        vec["v"].record(self.h.soma[0](0.5)._ref_v)
        vec["t"].record(self.h._ref_t)
    
        return vec



# Copyright 2014-2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.config.model.formats.json_util import JsonUtil

try:
    from ConfigParser import ConfigParser # Python 2
except:
    from configparser import ConfigParser # Python 3 #@UnresolvedImport

import argparse, os, io
import logging, logging.config as lc
from pkg_resources import resource_filename #@UnresolvedImport

class ApplicationConfig(object):
    ''' Convenience class that handles of application configuration
    from environment variables, .conf files and the command line
    using Python standard libraries and formats.
    '''
    
    _log = logging.getLogger(__name__)
    _DEFAULT_LOG_CONFIG = resource_filename(__name__, 'logging.conf')
    lc.fileConfig(_DEFAULT_LOG_CONFIG)
    
    def __init__(self, 
                 defaults,
                 name="app",
                 halp="Run application.",
                 default_log_config=None):
        self.application_name = name
        self.help = halp
        self.debug_enabled = False
    
        if default_log_config == None:
            default_log_config = ApplicationConfig._DEFAULT_LOG_CONFIG
        
        ApplicationConfig._log.info("default log config: %s" % (default_log_config))
        
        self.defaults = {
            'config_file_path': {
                'default': "%s.conf" % (self.application_name),
                'help': 'configuration file path'
            },
            'log_config_path': {
                'default': default_log_config,
                'help': 'logging configuration path'
            }
        }
        
        self.defaults.update(defaults)
        
        logging.info("defaults: %s" % (self.defaults))
        
        self.argparser = self.create_argparser()
                
        for key, value in self.defaults.items():
            setattr(self, key, value['default'])
    
    
    def load(self, command_line_args, disable_existing_loggers=True):
        ''' Load application configuration options, first from the environment,
        then from the configuration file, then from the command line.
            
        Each stage of loading can override the previous stage.
        
        Parameters
        ----------
        command_line_args : dict
            Parameters passed to the application.
        disable_existing_loggers : boolean
            Reset the logging system or not.
        
        Returns
        -------
        fileConfig
            Configuration object with all levels applied
        '''
        # read and apply options from the environment
        self.apply_configuration_from_environment()
        
        # command line so we can find the config file.
        parsed_args = self.parse_command_line_args(command_line_args)
        
        try:
            # read and apply the configuration file options
            config_file_path = parsed_args.config_file_path
            
            if config_file_path:
                self.config_file_path = config_file_path
            
            self.apply_configuration_from_file(self.config_file_path)
            
            # apply the remaining command line options
            self.apply_configuration_from_command_line(parsed_args)
        except Exception as e:
            ApplicationConfig._log.error("Could not load configuration file: %s\n%s" % 
                                         (parsed_args.config_file_path,
                                         e))
            raise
        
        if parsed_args.log_config_path:
            try:
                lc.fileConfig(self.log_config_path,
                              disable_existing_loggers=disable_existing_loggers)
            except:
                logging.error("Could not load log configuration file: %s" %
                             (parsed_args.log_config_path))
        else:
            # TODO: configure default logging
            pass
    
    
    def create_argparser(self):
        '''Initialization for the command-line parsing stage.
        
        An application specific prefix is applied to argument names. 
        
        Parameters
        ----------
        prog : string
            Application specific prefix for argument names.
        description : string
            A brief 'help' description of the application.
        
        Returns
        -------
        argParse.ArgumentParser
            The initialized argument parser object.
        
        Notes
        -----
        Defaults are set at the first environment reading.
        Command line args only override them when present
        '''
        parser = argparse.ArgumentParser(prog=self.application_name,
                                         description=self.help)
        for key, value in self.defaults.items():
            if key == 'config_file_path':
                parser.add_argument("%s" % (key), default=None, help=value['help'])
            else:
                parser.add_argument("--%s" % (key), default=None, help=value['help'])
       
        return parser
    
    
    def parse_command_line_args(self, args):
        '''Simply call the internal argparser object.
        
        Parameters
        ----------
        args : array
            Parameters passed to the application.
        
        Returns
        -------
        Namespace
            Parsed paramenters.
        '''
        return self.argparser.parse_args(args)
    
    
    def apply_configuration_from_command_line(self, parsed_args):
        '''Read application configuration variables from the command line.
        
        Unassigned variables are left unchanged if previously assigned,
        set to their default values,
        or None if no default is specified at init time.
        Assigned variables will overwrite the previous value.
        
        see: https://docs.python.org/2/howto/argparse.html
        
        Parameters
        ----------
        parsed_args : dict
            the arguments as parsed from the command line.
        
        '''
        logging.info('command_line args: %s' % (parsed_args))
        
        for key in self.defaults:
            parsed_value = getattr(parsed_args, key)
            if parsed_value and getattr(self, key) is None:
                setattr(self, key, parsed_value)
    
    
    def apply_configuration_from_environment(self):
        '''Read application configuration variables from the environment.
        
        The variable names are upper case and have a 
        prefix defined by the application.
            
        See: https://docs.python.org/2/library/os.html
        '''
        for key in self.defaults:
            environment_variable = "%s_%s" % (self.application_name.upper(), key.upper())
            environment_value = os.environ.get(environment_variable)
            if environment_value:
                setattr(self, key, environment_value)
    
    
    def from_json_file(self, json_path):
        '''Read an application configuration from a JSON format file.
        
        Parameters
        ----------
        json_path : string
            Path to the JSON file.
        
        Returns
        -------
        string
            An application configuration in INI format
        
        '''
        description = JsonUtil.read_json_file(json_path)
        
        return self.to_config_string(description)
    
    
    def from_json_string(self, json_string):
        '''Read a configuration from a JSON format string.
        
        Parameters
        ----------
        json_string : string
            A JSON-formatted string containing an application configuration.
        
        Returns
        -------
        string
            An application configuration in INI format
        '''
        description = JsonUtil.read_json_string(json_string)
        
        return self.to_config_string(description)
    
    
    def to_config_string(self, description):
        '''Create a configuration string from a dict.
        
        Parameters
        ----------
        description : dict
            Configuration options for an application.
        
        Returns
        -------
        string
            Equivalent configuration as an INI format string
        
        Notes
        -----
        The Python configparser library natively supports this functionality in Python 3.
        '''
        bps_config = description['biophys'][0]
        
        cfg_array = ['[biophys]']
        
        if 'log_config_path' in bps_config:
            cfg_array.append(str('log_config_path: %s' % bps_config['log_config_path']))
         
        if 'debug' in bps_config:
            cfg_array.append(str('debug: %s' % bps_config['debug']))
         
        if 'model_file' in bps_config:
            cfg_array.append(str('model_file: %s' % ','.join(bps_config['model_file'])))
            
        cfg_array.append("\n")
        
        bps_cfg_string = "\n".join(cfg_array)
        ApplicationConfig._log.info(bps_cfg_string)
        
        return bps_cfg_string
    
    
    def apply_configuration_from_file(self, config_file_path):
        ''' Read application configuration variables from a .conf file.
        
        Unassigned variables are set to their default values 
        or None if no default is specified at init time.
        The variables are found in a section named by the application.
        
        Parameters
        ----------
        config_file_path : string
            path to to an INI (.conf) or JSON format application config file.
        
        Returns
        -------
        
        see: https://docs.python.org/2/library/configparser.html
        '''
        none_defaults = {}
        
        # defaults are set in environment
        # they are only overriden by the config file if present
        for key in self.defaults:
            none_defaults[key] = None

        logging.info("none_defaults: %s" % (none_defaults))
            
        config = None
        
        try:
            config = ConfigParser(defaults=none_defaults, 
                                  allow_no_value=True)
        except:
            logging.warn("This python installation does not support configuration defaults.")
            config = ConfigParser()
        
        if config_file_path.endswith('.json'):
            cfg_string = ""
            cfg_string = self.from_json_file(config_file_path)
            config.readfp(io.BytesIO(cfg_string))
        else:
            config.read(config_file_path)
        
        for key in self.defaults:
            try:
                file_value = config.get(self.application_name, key)
                if file_value:
                    logging.info("setting %s to %s" % (key, file_value))
                    setattr(self, key, file_value)
            except:
                logging.info("Configuration option not specified: %s" %
                             (key))


# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from pkg_resources import resource_filename #@UnresolvedImport
import logging


class HocUtils(object):
    '''A helper class for containing references to NEUORN.
    
    Attributes
    ----------
    h : object
        The NEURON hoc object.
    nrn : object
        The NEURON python object.
    neuron : module
        The NEURON module.
    '''
    _log = logging.getLogger(__name__)
    h = None
    nrn = None
    neuron = None
    
    def __init__(self, description):
        import neuron
        import nrn
        
        self.h = neuron.h
        HocUtils.neuron = neuron
        HocUtils.nrn = nrn
        HocUtils.h = self.h
        
        self.description = description
        self.manifest = description.manifest
        
        self.hoc_files = description.data['neuron'][0]['hoc']
        
        self.initialize_hoc()
    
    
    def initialize_hoc(self):
        '''Basic setup for NEURON.'''
        h = self.h
        params = self.description.data['conditions'][0]
        
        for hoc_file in self.hoc_files:
            HocUtils._log.info("loading hoc file %s" % (hoc_file))
            HocUtils.h.load_file(str(hoc_file))
        
        h('starttime = startsw()')
        
        if 'celsius' in params:
            h.celsius = params['celsius']
            
        if 'v_init' in params:
            h.v_init = params['v_init']
            
        if 'dt' in params:
            h.dt = params['dt']
            h.steps_per_ms = 1.0/h.dt
            
        if 'tstop' in params:
            h.tstop = params['tstop']
            h.runStopAt = h.tstop